import numpy as np

from ase.utils.timing import Timer

from gpaw.utilities.lapack import inverse_general
from gpaw.transport.tools import get_matrix_index, collect_lead_mat, dot
import copy


class Banded_Sparse_HSD:
    # for lead's hamiltonian, overlap, and density matrix
    def __init__(self, dtype, ns, npk, index=None):
        self.band_index = index
        self.dtype = dtype
        self.H = []
        self.S = []
        self.D = []
        self.ns = ns
        self.npk = npk
        self.s = 0
        self.pk = 0
        for s in range(ns):
            self.H.append([])
            self.D.append([])
            for k in range(npk):
                self.H[s].append([])
                self.D[s].append([])
        for k in range(npk):
            self.S.append([])

    def reset(self, s, pk, mat, flag='S', init=False):
        assert mat.dtype == self.dtype
        if flag == 'S':
            spar = self.S
        elif flag == 'H':
            spar = self.H[s]
        elif flag == 'D':
            spar = self.D[s]
        if not init:
            spar[pk].reset(mat)
        elif self.band_index is not None:
            spar[pk] = Banded_Sparse_Matrix(self.dtype, mat, self.band_index)
        else:
            spar[pk] = Banded_Sparse_Matrix(self.dtype, mat)
            self.band_index = spar[pk].band_index
      
            
class Banded_Sparse_Matrix:
    def __init__(self, dtype, mat=None, band_index=None, tol=1e-9):
        self.tol = tol
        self.dtype = dtype
        self.band_index = band_index
        if mat is not None:
            if band_index is None:
                self.initialize(mat)
            else:
                self.reset(mat)
           
    def initialize(self, mat):
        # the indexing way needs mat[0][-1] = 0,otherwise will recover a
        # unsymmetric full matrix
        assert self.dtype == mat.dtype
        #assert mat[0][-1] < self.tol
        dim = mat.shape[-1]
        ku = -1
        kl = -1
        mat_sum = np.sum(abs(mat))
        spar_sum = 0
        while abs(mat_sum - spar_sum) > self.tol * 10:
            #ku += 1
            #kl += 1
            #ud_sum = 1
            #dd_sum = 1
            #while(ud_sum > self.tol):
            #    ku += 1
            #    ud_sum = np.sum(np.diag(abs(mat), ku))
            #while(dd_sum > self.tol):
            #    kl += 1
            #    dd_sum = np.sum(np.diag(abs(mat), -kl))
            #ku -= 1
            #kl -= 1
            ku = dim
            kl = dim
   
            # storage in the tranpose, bacause column major order for zgbsv_ function
            length = ((kl + ku + 1) * dim - kl * (kl + 1) // 2 -
                      ku * (ku + 1) // 2)
            
            self.spar = np.zeros([length], self.dtype)
                
            index1 = []
            index2 = []
            index0 = np.zeros((dim, 2 * kl + ku + 1), int)
           
            n = 0
            for i in range(kl, -1, -1):
                for j in range(dim - i):
                    index1.append(i + j)
                    index2.append(j)
                    index0[i + j, 2 * kl - i] = n
                    n += 1
        
            for i in range(1, ku + 1):
                for j in range(dim - i):
                    index1.append(j)
                    index2.append(j + i)
                    index0[j, 2 * kl + i] = n
                    n += 1
            
            index1 = np.array(index1)
            index2 = np.array(index2)
            
            self.band_index = (kl, ku, index0, index1, index2)
            self.spar = mat[index1, index2]
            spar_sum = np.sum(abs(self.recover()))

    def test1(self, n1, n2):
        index1 ,index2 = self.band_index[-2:]
        for i in range(len(index1)):
            if index1[i] == n1 and index2[i] == n2:
                print(i)
                
    def recover(self):
        index0, index1, index2 = self.band_index[-3:]
        dim = index0.shape[0]
        mat = np.zeros([dim, dim], self.dtype)
        mat[index1, index2] = self.spar
        return mat
 
    def reset(self, mat):
        index1, index2 = self.band_index[-2:]
        assert self.dtype == mat.dtype
        self.spar = mat[index1, index2]

    def reset_from_others(self, bds_mm1, bds_mm2, c1, c2):
        assert self.dtype == complex
        self.spar = c1 * bds_mm1.spar + c2 * bds_mm2.spar
            
    def reset_minus(self, mat, full=False):
        assert self.dtype == complex
        index1, index2 = self.band_index[-2:]
        if full:
            self.spar -= mat[index1, index2]
        else:
            self.spar -= mat.recover()[index1, index2]
    
    def reset_plus(self, mat, full=False):
        assert self.dtype == complex
        index1, index2 = self.band_index[-2:]
        if full:
            self.spar += mat[index1, index2]
        else:
            self.spar += mat.recover()[index1, index2]

    def test_inv_speed(self):
        full_mat = self.recover()
        timer = Timer()
        timer.start('full_numpy')
        np.linalg.inv(full_mat)
        timer.stop('full_numpy')
        
        timer.start('full_lapack')
        inverse_general(full_mat)
        timer.stop('full_lapack')
        
        timer.start('sparse_lapack')
        self.inv()
        timer.stop('sparse_lapack')
        
        times = []
        methods = ['full_numpy', 'full_lapack', 'sparse_lapack']
        for name in methods:
            time = timer.timers[name,]
            print(name, time)
            times.append(time)
        
        mintime = np.min(times)
        self.inv_method = methods[np.argmin(times)]
        print('mintime', mintime)
                
    def inv(self):
        #kl, ku, index0 = self.band_index[:3]
        #dim = index0.shape[0]
        #inv_mat = np.eye(dim, dtype=complex)
        #ldab = 2*kl + ku + 1
        #source_mat = self.spar[index0]
        #assert source_mat.flags.contiguous
        #info = _gpaw.linear_solve_band(source_mat, inv_mat, kl, ku)
        #return inv_mat
        return np.linalg.inv(self.recover()).copy()
       
class Tp_Sparse_HSD:
    def __init__(self, dtype, ns, npk, ll_index, ex=True):
        self.dtype = dtype
        self.ll_index = ll_index
        self.extended = ex
        self.H = []
        self.S = []
        self.D = []
        self.G = []
        self.ns = ns
        self.npk = npk
        self.s = 0
        self.pk = 0
        self.band_indices = None
        for s in range(ns):
            self.H.append([])
            self.D.append([])
            for k in range(npk):
                self.H[s].append([])
                self.D[s].append([])
        for k in range(npk):
            self.S.append([])
        self.G = Tp_Sparse_Matrix(complex, self.ll_index,
                                                    None, None, self.extended)
    
    def reset(self, s, pk, mat, flag='S', init=False):
        if flag == 'S':
            spar = self.S
        elif flag == 'H':
            spar = self.H[s]
        elif flag == 'D':
            spar = self.D[s]
        if not init:
            spar[pk].reset(mat)
        elif self.band_indices is None:
            spar[pk] = Tp_Sparse_Matrix(self.dtype, self.ll_index, mat,
                                                          None, self.extended)
            self.band_indices = spar[pk].band_indices
        else:
            spar[pk] = Tp_Sparse_Matrix(self.dtype, self.ll_index, mat,
                                             self.band_indices, self.extended)

    def append_lead_as_buffer(self, lead_hsd, lead_couple_hsd, ex_index, tp=None):
        assert self.extended == True
        clm = collect_lead_mat
        if tp is not None:
            tp.log('append_lead_as_buffer(), npk : {0}  ns : {1}'.format(self.npk, self.ns))
        for pk in range(self.npk):
            if tp is not None:
                tp.log('append_lead_as_buffer(), pk : {0}'.format(pk))
            diag_h, upc_h, dwnc_h = clm(lead_hsd, lead_couple_hsd, 0, pk)
            self.S[pk].append_ex_mat(diag_h, upc_h, dwnc_h, ex_index)
            for s in range(self.ns):
                if tp is not None:
                    tp.log('append_lead_as_buffer(), s : {0}'.format(s))
                    tp.log('    clm()')
                diag_h, upc_h, dwnc_h = clm(lead_hsd,
                                                  lead_couple_hsd, s, pk, 'H')
                if tp is not None:
                    tp.log('    append_ex_mat()')
                self.H[s][pk].append_ex_mat(diag_h, upc_h, dwnc_h, ex_index)
                if tp is not None:
                    tp.log('    clm()')
                diag_h, upc_h, dwnc_h = clm(lead_hsd,
                                                  lead_couple_hsd, s, pk, 'D')
                if tp is not None:
                    tp.log('    append_ex_mat()')
                self.D[s][pk].append_ex_mat(diag_h, upc_h, dwnc_h, ex_index, tp=tp)
  
    def calculate_eq_green_function(self, zp, sigma, ex=True, full=False):
        s, pk = self.s, self.pk
        #print('calculate_eq_green_function() s: {0} pk : {1}'.format(s,pk))
        self.G.reset_from_others(self.S[pk], self.H[s][pk], zp, -1, init=True)
        self.G.substract_sigma(sigma)
        if full:
            return np.linalg.inv(self.G.recover())
        else:
            #self.G.test_inv_eq()
            self.G.inv_eq()
            return self.G.recover(ex)

    def calculate_ne_green_function(self, zp, sigma, ffocc, ffvir, ex=True):
        s, pk = self.s, self.pk
        self.G.reset_from_others(self.S[pk], self.H[s][pk], zp, -1)
        self.G.substract_sigma(sigma)
        gammaocc = []
        gammavir = []
        for ff0, ff1, tgt in zip(ffocc, ffvir, sigma):
            full_tgt = tgt.recover()
            gammaocc.append(ff0 * 1.j * (full_tgt - full_tgt.T.conj()))
            gammavir.append(ff1 * 1.j * (full_tgt - full_tgt.T.conj()))
        glesser, ggreater = self.G.calculate_non_equilibrium_green(gammaocc,
                                                                gammavir, ex)
        return glesser, ggreater

    def abstract_sub_green_matrix(self, zp, sigma, l1, l2, inv_mat=None):
        if inv_mat is None:
            s, pk = self.s, self.pk
            self.G.reset_from_others(self.S[pk], self.H[s][pk], zp, -1)
            self.G.substract_sigma(sigma)
            inv_mat = self.G.inv_ne()
            gr_sub = inv_mat[l2][l1][-1]
            return gr_sub, inv_mat
        else:
            gr_sub = inv_mat[l2][l1][-1]
            return gr_sub
       
class Tp_Sparse_Matrix:
    def __init__(self, dtype, ll_index, mat=None, band_indices=None, ex=True):
    # ll_index : lead_layer_index
    # matrix stored here will be changed to inversion

        self.lead_num = len(ll_index)
        self.ll_index = ll_index
        self.ex_ll_index = copy.deepcopy(ll_index[:])
        self.extended = ex
        self.dtype = dtype
        self.initialize()
        self.band_indices = band_indices
        if self.band_indices is None:
            self.initialize_band_indices()
        if mat is not None:
            self.reset(mat, True)
        
    def initialize_band_indices(self):
        self.band_indices = [None]
        for i in range(self.lead_num):
            self.band_indices.append([])
            for j in range(self.ex_lead_nlayer[i] - 1):
                self.band_indices[i + 1].append(None)
        
    def initialize(self):
    # diag_h : diagonal lead_hamiltonian
    # upc_h : superdiagonal lead hamiltonian
    # dwnc_h : subdiagonal lead hamiltonian
        self.diag_h = []
        self.upc_h = []
        self.dwnc_h = []
        self.lead_nlayer = []
        self.ex_lead_nlayer = []
        self.mol_index = self.ll_index[0][0]
        self.nl = 1
        self.nb = len(self.mol_index)
        self.length = self.nb * self.nb
        self.mol_h = []

        for i in range(self.lead_num):
            self.diag_h.append([])
            self.upc_h.append([])
            self.dwnc_h.append([])
            self.lead_nlayer.append(len(self.ll_index[i]))
            if self.extended:
                self.ex_lead_nlayer.append(len(self.ll_index[i]) + 1)
            else:
                self.ex_lead_nlayer.append(len(self.ll_index[i]))
            
            assert (self.ll_index[i][0] == self.mol_index).all()
            self.nl += self.lead_nlayer[i] - 1
            
            for j in range(self.lead_nlayer[i] - 1):
                self.diag_h[i].append([])
                self.upc_h[i].append([])
                self.dwnc_h[i].append([])
                len1 = len(self.ll_index[i][j])
                len2 = len(self.ll_index[i][j + 1])
                self.length += 2 * len1 * len2 + len2 * len2
                self.nb += len2
            
            if self.extended:
                self.diag_h[i].append([])
                self.upc_h[i].append([])
                self.dwnc_h[i].append([])
        self.ex_nb = self.nb

    def append_ex_mat(self, diag_h, upc_h, dwnc_h, ex_index, tp=None):
        assert self.extended
        if tp is not None:
            tp.log('        append_ex_mat()')
        for i in range(self.lead_num):
            if tp is not None:
                tp.log('        append_ex_mat() i: {0}'.format(i))
            self.diag_h[i][-1] = diag_h[i]
            if tp is not None:
                tp.log('        append_ex_mat() diag_h')
            self.upc_h[i][-1] = upc_h[i]
            if tp is not None:
                tp.log('        append_ex_mat() upc_h')
            self.dwnc_h[i][-1] = dwnc_h[i]
            if tp is not None:
                tp.log('        append_ex_mat() dwnc_h')
            self.ex_ll_index[i].append(ex_index[i])
            if tp is not None:
                tp.log('        append_ex_mat() .append')
            self.ex_nb += len(ex_index[i])
            if tp is not None:
                tp.log('        append_ex_mat() += len()')
            
  
    def abstract_layer_info(self):
        self.basis_to_layer = np.empty([self.nb], int)
        self.neighbour_layers = np.zeros([self.nl, self.lead_num], int) - 1

        for i in self.mol_index:
            self.basis_to_layer[i] = 0
        nl = 1
        
        for i in range(self.lead_num):
            for j in range(self.lead_nlayer[i] - 1):
                for k in self.ll_index[i][j]:
                    self.basis_to_layer[k] = nl
                nl += 1

        nl = 1
        for i in range(self.lead_num):
            self.neighbour_layers[0][i] = nl
            first = nl
            for j in range(self.lead_nlayer[i] - 1):
                if nl == first:
                    self.neighbour_layers[nl][0] = 0
                    if j != self.lead_nlayer[i] - 2:
                        self.neighbour_layers[nl][1] = nl + 1
                else:
                    self.neighbour_layers[nl][0] = nl - 1
                    if j != self.lead_nlayer[i] - 2:
                        self.neighbour_layers[nl][1] = nl + 1
                nl += 1
              
    def reset(self, mat, init=False):
        assert mat.dtype == self.dtype
        ind = get_matrix_index(self.mol_index)
        if init:
            self.mol_h = Banded_Sparse_Matrix(self.dtype, mat[ind.T, ind],
                                               self.band_indices[0])
            if self.band_indices[0] is None:
                self.band_indices[0] = self.mol_h.band_index
        else:
            self.mol_h.reset(mat[ind.T, ind])

        for i in range(self.lead_num):
            for j in range(self.lead_nlayer[i] - 1):
                ind = get_matrix_index(self.ll_index[i][j])
                ind1 = get_matrix_index(self.ll_index[i][j + 1])
                indr1, indc1 = get_matrix_index(self.ll_index[i][j],
                                                      self.ll_index[i][j + 1])
                indr2, indc2 = get_matrix_index(self.ll_index[i][j + 1],
                                                  self.ll_index[i][j])
                if init:
                    self.diag_h[i][j] = Banded_Sparse_Matrix(self.dtype,
                                                             mat[ind1.T, ind1],
                                                 self.band_indices[i + 1][j])
                    if self.band_indices[i + 1][j] is None:
                        self.band_indices[i + 1][j] = \
                                                  self.diag_h[i][j].band_index
                else:
                    self.diag_h[i][j].reset(mat[ind1.T, ind1])
                self.upc_h[i][j] = mat[indr1, indc1]
                self.dwnc_h[i][j] = mat[indr2, indc2]
       
    def reset_from_others(self, tps_mm1, tps_mm2, c1, c2, init=False):
        #self.mol_h = c1 * tps_mm1.mol_h + c2 * tps_mm2.mol_h
        #print('reset_from_others {0}  {1}'.format(tps_mm1, tps_mm2))
        if init:
            self.mol_h = Banded_Sparse_Matrix(complex)
        
        self.mol_h.spar = c1 * tps_mm1.mol_h.spar + c2 * tps_mm2.mol_h.spar
        self.mol_h.band_index = tps_mm1.mol_h.band_index
        self.ex_lead_nlayer = tps_mm1.ex_lead_nlayer
        self.ex_ll_index = tps_mm1.ex_ll_index
        self.ex_nb = tps_mm1.ex_nb
        
        for i in range(self.lead_num):
            for j in range(self.ex_lead_nlayer[i] - 1):
                assert (tps_mm1.ex_ll_index[i][j] == tps_mm2.ex_ll_index[i][j]).all()
                if init:
                    self.diag_h[i][j] = Banded_Sparse_Matrix(complex)
                    self.diag_h[i][j].band_index = \
                                             tps_mm1.diag_h[i][j].band_index
                
                self.diag_h[i][j].spar = c1 * tps_mm1.diag_h[i][j].spar + \
                                      c2 * tps_mm2.diag_h[i][j].spar
                self.upc_h[i][j] = c1 * tps_mm1.upc_h[i][j] + \
                                      c2 * tps_mm2.upc_h[i][j]
                self.dwnc_h[i][j] = c1 * tps_mm1.dwnc_h[i][j] + \
                                      c2 * tps_mm2.dwnc_h[i][j]
  
    def substract_sigma(self, sigma):
        if self.extended:
            n = -2
        else:
            n = -1
        for i in range(self.lead_num):
            self.diag_h[i][n].reset_minus(sigma[i])
        
    def recover(self, ex=False):
        if ex:
            nb = self.ex_nb
            lead_nlayer = self.ex_lead_nlayer
            ll_index = self.ex_ll_index
        else:
            nb = self.nb
            lead_nlayer = self.lead_nlayer
            ll_index = self.ll_index
        
        mat = np.zeros([nb, nb], self.dtype)
        ind = get_matrix_index(ll_index[0][0])
        
        mat[ind.T, ind] = self.mol_h.recover()
        
        gmi = get_matrix_index
        for i in range(self.lead_num):
            for j in range(lead_nlayer[i] - 1):
                ind = gmi(ll_index[i][j])
                ind1 = gmi(ll_index[i][j + 1])
                indr1, indc1 = gmi(ll_index[i][j], ll_index[i][j + 1])
                indr2, indc2 = gmi(ll_index[i][j + 1], ll_index[i][j])
                mat[ind1.T, ind1] = self.diag_h[i][j].recover()
                mat[indr1, indc1] = self.upc_h[i][j]
                mat[indr2, indc2] = self.dwnc_h[i][j]
        return mat

    def test_inv_eq(self, tol=1e-9):
        tp_mat = copy.deepcopy(self)
        tp_mat.inv_eq()
        mol_h = dot(tp_mat.mol_h.recover(), self.mol_h.recover())
        for i in range(self.lead_num):
            mol_h += dot(tp_mat.upc_h[i][0], self.dwnc_h[i][0])
        diff = np.max(abs(mol_h - np.eye(mol_h.shape[0])))
        if diff > tol:
            print('warning, mol_diff', diff)
        for i in range(self.lead_num):
            for j in range(self.lead_nlayer[i] - 2):
                diag_h = dot(tp_mat.diag_h[i][j].recover(),
                                                  self.diag_h[i][j].recover())
                diag_h += dot(tp_mat.dwn_h[i][j], self.upc_h[i][j])
                diag_h += dot(tp_mat.upc_h[i][j + 1], self.dwnc_h[i][j + 1])
                diff = np.max(abs(diag_h - np.eye(diag_h.shape[0])))
                if diff > tol:
                    print('warning, diag_diff', i, j, diff)
            j = self.lead_nlayer[i] - 2
            diag_h = dot(tp_mat.diag_h[i][j].recover(),
                                                  self.diag_h[i][j].recover())
            diag_h += dot(tp_mat.dwnc_h[i][j], self.upc_h[i][j])
            diff = np.max(abs(diag_h - np.eye(diag_h.shape[0])))
            if diff > tol:
                print('warning, diag_diff', i, j, diff)
                                                
    def inv_eq(self):
        q_mat = []
        for i in range(self.lead_num):
            q_mat.append([])
            nll = self.lead_nlayer[i]
            for j in range(nll - 1):
                q_mat[i].append([])
            end = nll - 2
            q_mat[i][end] =  self.diag_h[i][end].inv()
          
            for j in range(end - 1, -1, -1):
                self.diag_h[i][j].reset_minus(self.dotdot(
                                                    self.upc_h[i][j + 1],
                                                         q_mat[i][j + 1],
                                            self.dwnc_h[i][j + 1]), full=True)
                q_mat[i][j] = self.diag_h[i][j].inv()
        h_mm = self.mol_h

        for i in range(self.lead_num):
            h_mm.reset_minus(self.dotdot(self.upc_h[i][0], q_mat[i][0],
                                                self.dwnc_h[i][0]), full=True)
        inv_h_mm = h_mm.inv()
        h_mm.reset(inv_h_mm)
        
        for i in range(self.lead_num):
            tmp_dc = self.dwnc_h[i][0].copy()
            self.dwnc_h[i][0] = -self.dotdot(q_mat[i][0], tmp_dc, inv_h_mm)
            self.upc_h[i][0] = -self.dotdot(inv_h_mm, self.upc_h[i][0],
                                                                    q_mat[i][0])
            dim = len(self.ll_index[i][1])
            self.diag_h[i][0].reset(dot(q_mat[i][0], np.eye(dim) -
                                                  dot(tmp_dc, self.upc_h[i][0])))
            for j in range(1, self.lead_nlayer[i] - 1):
                tmp_dc = self.dwnc_h[i][j].copy()
                self.dwnc_h[i][j] = -self.dotdot(q_mat[i][j], tmp_dc,
                                                self.diag_h[i][j - 1].recover())
                self.upc_h[i][j] = -self.dotdot(self.diag_h[i][j - 1].recover(),
                                                    self.upc_h[i][j],
                                                     q_mat[i][j])
                dim = len(self.ll_index[i][j + 1])
                self.diag_h[i][j].reset(dot(q_mat[i][j], np.eye(dim) -
                                           dot(tmp_dc, self.upc_h[i][j])))

    def inv_ne(self):
        q_mat = []
        qi_mat = []
        inv_mat = []
        #structure of inv_mat inv_cols_1, inv_cols_2, ..., inv_cols_n (n:lead_num)
        #structure of inv_cols_i   inv_cols_l1, inv_cols_l2,..., inv_cols_ln, inv_cols_mm(matrix)
        #structure of inv_cols_li  inv_cols_ll1, inv_cols_ll2,...,inv_cols_ll3
        for i in range(self.lead_num):
            q_mat.append([])
            qi_mat.append([])
            inv_mat.append([])
            
            nll = self.lead_nlayer[i]
            for j in range(nll - 1):
                q_mat[i].append([])
                qi_mat[i].append([])
                
            for j in range(self.lead_num):
                inv_mat[i].append([])
                nll_j = self.lead_nlayer[j]
                for k in range(nll_j - 1):
                    inv_mat[i][j].append([])
            inv_mat[i].append([])
            
            end = nll - 2
            q_mat[i][end] =  self.diag_h[i][end].inv()
            for j in range(end - 1, -1, -1):
                tmp_diag_h = copy.deepcopy(self.diag_h[i][j])
                tmp_diag_h.reset_minus(self.dotdot(self.upc_h[i][j + 1],
                                                     q_mat[i][j + 1],
                                                  self.dwnc_h[i][j + 1]),
                                        full=True)
                q_mat[i][j] = tmp_diag_h.inv()
        # above get all the q matrix, then if want to solve the cols
        # cooresponding to the lead i, the q_mat[i] will not be used

        #q_mm = self.mol_h.recover()
        q_mm = copy.deepcopy(self.mol_h)
        for i in range(self.lead_num):
            #q_mm -= dot(dot(self.upc_h[i][0], q_mat[i][0]),
            #                                 self.dwnc_h[i][0])
            q_mm.reset_minus(self.dotdot(self.upc_h[i][0],
                                  q_mat[i][0], self.dwnc_h[i][0]), full=True)
        
        for i in range(self.lead_num):
        # solve the corresponding cols to the lead i
            nll = self.lead_nlayer[i]
    
            #qi_mat[i][0] = q_mm + self.dotdot(self.upc_h[i][0],q_mat[i][0],
            #                                                self.dwnc_h[i][0])
            q_mm_tmp = copy.deepcopy(q_mm)
            q_mm_tmp.reset_plus(self.dotdot(self.upc_h[i][0],q_mat[i][0],
                                                self.dwnc_h[i][0]), full=True)
            
            #inv(qi_mat[i][0])
            qi_mat[i][0] = q_mm_tmp.inv()
            for j in range(1, nll - 1):
                tmp_diag_h = copy.deepcopy(self.diag_h[i][j - 1])
                tmp_diag_h.reset_minus(self.dotdot(self.dwnc_h[i][j -1],
                                                        qi_mat[i][j - 1],
                                                        self.upc_h[i][j - 1]),
                                                       full=True)
                qi_mat[i][j] = tmp_diag_h.inv()

            tmp_diag_h = copy.deepcopy(self.diag_h[i][nll - 2])
            tmp_diag_h.reset_minus(self.dotdot(self.dwnc_h[i][nll - 2],
                                                qi_mat[i][nll -2],
                                               self.upc_h[i][nll -2]),
                                                  full=True)
            inv_mat[i][i][nll - 2] = tmp_diag_h.inv()
            
            for j in range(nll - 3, -1, -1):
                inv_mat[i][i][j] = -self.dotdot(qi_mat[i][j + 1],
                                                  self.upc_h[i][j + 1],
                                                 inv_mat[i][i][j + 1])
            inv_mat[i][self.lead_num] = -self.dotdot(qi_mat[i][0],
                                                  self.upc_h[i][0],
                                                  inv_mat[i][i][0])
            
            for j in range(self.lead_num):
                if j != i:
                    nlj = self.lead_nlayer[j]
                    inv_mat[i][j][0] = -self.dotdot(q_mat[j][0], self.dwnc_h[j][0],
                                                inv_mat[i][self.lead_num])
                    for k in range(1, nlj - 1):
                        inv_mat[i][j][k] = -self.dotdot(q_mat[j][k], self.dwnc_h[j][k],
                                                inv_mat[i][j][k - 1])
        return inv_mat
  
  
    def combine_inv_mat(self, inv_mat):
        nb = self.nb
        mat = np.zeros([nb, nb], complex)
        for i in range(self.lead_num):
            indr, indc = get_matrix_index(self.ll_index[i][0],
                                          self.ll_index[i][-1])
            mat[indr, indc] = inv_mat[i][self.lead_num]
            for j in range(self.lead_num):
                for k in range(1, self.lead_nlayer[j]):
                    indr, indc = get_matrix_index(self.ll_index[j][k],
                                                  self.ll_index[i][-1])
                    mat[indr, indc] = inv_mat[i][j][k - 1]
        return mat
  
    def dotdot(self, mat1, mat2, mat3):
        return dot(mat1, dot(mat2, mat3))
    
    def calculate_non_equilibrium_green(self, se_less, se_great, ex=True):
        inv_mat = self.inv_ne()
        glesser = self.calculate_keldysh_green(inv_mat, se_less, ex)
        ggreater = self.calculate_keldysh_green(inv_mat, se_great, ex)
        return glesser, ggreater
    
    def calculate_keldysh_green(self, inv_mat, keldysh_se, ex=True):
        #se_less less selfenergy, structure  se_1, se_2, se_3,..., se_n
        #the lead sequence of se_less should be the same to self.ll_index
        self.mol_h.spar.fill(0.0)
        for i in range(self.lead_num):
            nll = self.lead_nlayer[i]
            for j in range(nll - 1):
                self.diag_h[i][j].spar.fill(0.0)
                self.upc_h[i][j].fill(0.0)
                self.dwnc_h[i][j].fill(0.0)
        
        for i in range(self.lead_num):
            # less selfenergy loop
            self.mol_h.reset_plus(self.dotdot(inv_mat[i][self.lead_num],
                                              keldysh_se[i],
                                          inv_mat[i][self.lead_num].T.conj()),
                                              full=True)
            for j in range(self.lead_num):
               # matrix operation loop
                nlj = self.lead_nlayer[j]
                self.diag_h[j][0].reset_plus(self.dotdot(inv_mat[i][j][0],
                                                    keldysh_se[i],
                                                 inv_mat[i][j][0].T.conj()),
                                                      full=True)
            
                self.dwnc_h[j][0] += self.dotdot(inv_mat[i][j][0], keldysh_se[i],
                                           inv_mat[i][self.lead_num].T.conj())
            
                self.upc_h[j][0] += self.dotdot(inv_mat[i][self.lead_num],
                                                keldysh_se[i],
                                            inv_mat[i][j][0].T.conj())
            
                for k in range(1, nlj -1):
                    self.diag_h[j][k].reset_plus(self.dotdot(inv_mat[i][j][k],
                                                             keldysh_se[i],
                                                 inv_mat[i][j][k].T.conj()),
                                                    full=True)
                    
                    self.dwnc_h[j][k] += self.dotdot(inv_mat[i][j][k],
                                                     keldysh_se[i],
                                                inv_mat[i][j][k - 1].T.conj())
                        
                    self.upc_h[j][k] +=  self.dotdot(inv_mat[i][j][k - 1],
                                                     keldysh_se[i],
                                                    inv_mat[i][j][k].T.conj())
        return self.recover(ex)

    def test_inv_speed(self):
        full_mat = self.recover()
        timer = Timer()
        timer.start('full_numpy')
        np.linalg.inv(full_mat)
        timer.stop('full_numpy')
        
        timer.start('full_lapack')
        inverse_general(full_mat)
        timer.stop('full_lapack')
        
        timer.start('sparse_lapack')
        self.inv_eq()
        timer.stop('sparse_lapack')
        
        timer.start('sparse_lapack_ne')
        self.inv_ne()
        timer.stop('sparse_lapack_ne')
        
        times = []
        methods = ['full_numpy', 'full_lapack', 'sparse_lapack']
        for name in methods:
            time = timer.timers[name,]
            print(name, time)
            times.append(time)
        
        mintime = np.min(times)
        self.inv_method = methods[np.argmin(times)]
        print('mintime', mintime)
        
        print('sparse_lapack_ne', timer.timers['sparse_lapack_ne',])

class CP_Sparse_HSD:
    def __init__(self, dtype, ns, npk, index=None):
        self.index = index
        self.dtype = dtype
        self.H = []
        self.S = []
        self.D = []
        self.ns = ns
        self.npk = npk
        self.s = 0
        self.pk = 0
        for s in range(ns):
            self.H.append([])
            self.D.append([])
            for k in range(npk):
                self.H[s].append([])
                self.D[s].append([])
        for k in range(npk):
            self.S.append([])

    def reset(self, s, pk, mat, flag='S', init=False):
        assert mat.dtype == self.dtype
        if flag == 'S':
            spar = self.S
        elif flag == 'H':
            spar = self.H[s]
        elif flag == 'D':
            spar = self.D[s]
        if not init:
            spar[pk].reset(mat)
        elif self.index is not None:
            spar[pk] = CP_Sparse_Matrix(self.dtype, mat, self.index)
        else:
            spar[pk] = CP_Sparse_Matrix(self.dtype, mat)
            self.index = spar[pk].index
     
class CP_Sparse_Matrix:
    def __init__(self, dtype, mat=None, index=None, flag=None, tol=1e-9):
        self.tol = tol
        self.index = index
        self.dtype = dtype
        self.flag = flag
        if mat is not None:
            if self.index is None:
                self.initialize(mat)
            else:
                self.reset(mat)
        
    def initialize(self, mat):
        assert self.dtype == mat.dtype
        dim = mat.shape[-1]
        ud_array = np.empty([dim])
        dd_array = np.empty([dim])
        for i in range(dim):
            ud_array[i] = np.sum(abs(np.diag(mat, i)))
            dd_array[i] = np.sum(abs(np.diag(mat, -i)))
        spar_sum = 0
        mat_sum = np.sum(abs(mat))
        if np.sum(abs(ud_array)) >  np.sum(abs(dd_array)):
            self.flag = 'U'
            i = -1
            while abs(mat_sum - spar_sum) > self.tol * 10:
                i += 1
                while ud_array[i] < self.tol and  i < dim - 1:
                    i += 1
                self.index = (i, dim)
                ldab = dim - i
                self.spar = mat[:ldab, i:].copy()
                spar_sum = np.sum(abs(self.spar))
        else:
            self.flag = 'L'
            i = -1
            while abs(mat_sum - spar_sum) > self.tol * 10:
                i += 1
                while dd_array[i] < self.tol and  i < dim - 1:
                    i += 1
                self.index = (-i, dim)
                ldab = dim - i
                self.spar = mat[i:, :ldab].copy()
                spar_sum = np.sum(abs(self.spar))
        
    def reset(self, mat):
        assert mat.dtype == self.dtype and mat.shape[-1] == self.index[1]
        dim = mat.shape[-1]
        if self.index[0] > 0:
            ldab = dim - self.index[0]
            self.spar = mat[:ldab, self.index[0]:].copy()
        else:
            ldab = dim + self.index[0]
            self.spar = mat[-self.index[0]:, :ldab].copy()

    def recover(self, trans='n'):
        nb = self.index[1]
        mat = np.zeros([nb, nb], self.dtype)
        if self.index[0] > 0:
            ldab = nb - self.index[0]
            mat[:ldab, self.index[0]:] = self.spar
        else:
            ldab = nb + self.index[0]
            mat[-self.index[0]:, :ldab] = self.spar
        if trans == 'c':
            if self.dtype == float:
                mat = mat.T.copy()
            else:
                mat = mat.T.conj()
        return mat

class Se_Sparse_Matrix:
    def __init__(self, mat, tri_type, nn=None, tol=1e-9):
        # coupling sparse matrix A_ij!=0 if i>dim-nn and j>dim-nn (for right selfenergy)
        # or A_ij!=0 if i<nn and j<nn (for left selfenergy, dim is the shape of A)
        self.tri_type = tri_type
        self.tol = tol
        self.nb = mat.shape[-1]
        self.spar = []
        if nn is None:
            self.initialize(mat)
        else:
            self.reset(mat, nn)

    def initialize(self, mat):
        self.nn = 0
        nb = self.nb
        tol = self.tol
        if self.tri_type == 'L':
            while self.nn < nb and np.sum(abs(mat[self.nn])) > tol:
                self.nn += 1
            self.spar = mat[:self.nn, :self.nn].copy()
        else:
            while self.nn < nb and np.sum(abs(mat[nb - self.nn - 1])) > tol:
                self.nn += 1
            self.spar = mat[-self.nn:, -self.nn:].copy()
        
        diff = abs(np.sum(abs(mat)) - np.sum(abs(self.spar)))
        if diff > tol * 10:
            print('Warning! Sparse Matrix Diff', diff)
        
    def reset(self, mat, nn=None):
        if nn is not None:
            self.nn = nn
        if self.tri_type == 'L':
            self.spar = mat[:self.nn, :self.nn].copy()
        else:
            self.spar = mat[-self.nn:, -self.nn:].copy()
  
    def restore(self):
        mat = np.zeros([self.nb, self.nb], complex)
        if self.tri_type == 'L':
            mat[:self.nn, :self.nn] = self.spar
        else:
            mat[-self.nn:, -self.nn:] = self.spar
        return mat



"""Module defining  ``Eigensolver`` classes."""

import numpy as np

from gpaw.utilities.blas import axpy
from gpaw.eigensolvers.eigensolver import Eigensolver

from gpaw import extra_parameters

class RMM_DIIS(Eigensolver):
    """RMM-DIIS eigensolver

    It is expected that the trial wave functions are orthonormal
    and the integrals of projector functions and wave functions
    ``nucleus.P_uni`` are already calculated

    Solution steps are:

    * Subspace diagonalization
    * Calculation of residuals
    * Improvement of wave functions:  psi' = psi + lambda PR + lambda PR'
    * Orthonormalization"""

    def __init__(self, keep_htpsit=True, blocksize=10,
                 fixed_trial_step=None):
        self.fixed_trial_step = fixed_trial_step
        Eigensolver.__init__(self, keep_htpsit, blocksize)

    def iterate_one_k_point(self, hamiltonian, wfs, kpt):
        """Do a single RMM-DIIS iteration for the kpoint"""

        psit_nG, R_nG = self.subspace_diagonalize(hamiltonian, wfs, kpt)

        self.timer.start('RMM-DIIS')
        if self.keep_htpsit:
            self.calculate_residuals(kpt, wfs, hamiltonian, psit_nG,
                                     kpt.P_ani, kpt.eps_n, R_nG)

        def integrate(a_G, b_G):
            return np.real(wfs.integrate(a_G, b_G, global_integral=False))

        comm = wfs.gd.comm
        B = self.blocksize
        dR_xG = wfs.empty(B, q=kpt.q)
        P_axi = wfs.pt.dict(B)
        error = 0.0
        for n1 in range(0, wfs.bd.mynbands, B):
            n2 = n1 + B
            if n2 > wfs.bd.mynbands:
                n2 = wfs.bd.mynbands
                B = n2 - n1
                P_axi = dict((a, P_xi[:B]) for a, P_xi in P_axi.items())
                dR_xG = dR_xG[:B]
                
            n_x = np.arange(n1, n2)
            psit_xG = psit_nG[n1:n2]
            
            if self.keep_htpsit:
                R_xG = R_nG[n1:n2]
            else:
                R_xG = wfs.empty(B, q=kpt.q)
                wfs.apply_pseudo_hamiltonian(kpt, hamiltonian, psit_xG, R_xG)
                wfs.pt.integrate(psit_xG, P_axi, kpt.q)
                self.calculate_residuals(kpt, wfs, hamiltonian, psit_xG,
                                         P_axi, kpt.eps_n[n_x], R_xG, n_x)

            for n in n_x:
                if kpt.f_n is None:
                    weight = kpt.weight
                else:
                    weight = kpt.f_n[n]
                if self.nbands_converge != 'occupied':
                    if wfs.bd.global_index(n) < self.nbands_converge:
                        weight = kpt.weight
                    else:
                        weight = 0.0
                error += weight * integrate(R_xG[n - n1], R_xG[n - n1])

            # Precondition the residual:
            self.timer.start('precondition')
            ekin_x = self.preconditioner.calculate_kinetic_energy(
                psit_xG, kpt)
            dpsit_xG = self.preconditioner(R_xG, kpt, ekin_x)
            self.timer.stop('precondition')

            # Calculate the residual of dpsit_G, dR_G = (H - e S) dpsit_G:
            wfs.apply_pseudo_hamiltonian(kpt, hamiltonian, dpsit_xG, dR_xG)
            self.timer.start('projections')
            wfs.pt.integrate(dpsit_xG, P_axi, kpt.q)
            self.timer.stop('projections')
            self.calculate_residuals(kpt, wfs, hamiltonian, dpsit_xG,
                                     P_axi, kpt.eps_n[n_x], dR_xG, n_x,
                                     calculate_change=True)
            
            # Find lam that minimizes the norm of R'_G = R_G + lam dR_G
            RdR_x = np.array([integrate(dR_G, R_G)
                              for R_G, dR_G in zip(R_xG, dR_xG)])
            dRdR_x = np.array([integrate(dR_G, dR_G) for dR_G in dR_xG])
            comm.sum(RdR_x)
            comm.sum(dRdR_x)

            lam_x = -RdR_x / dRdR_x
            if extra_parameters.get('PK', False):
                lam_x[:] = np.where(lam_x>0.0, lam_x, 0.2)
            # Calculate new psi'_G = psi_G + lam pR_G + lam2 pR'_G
            #                      = psi_G + p((lam+lam2) R_G + lam*lam2 dR_G)
            for lam, R_G, dR_G in zip(lam_x, R_xG, dR_xG):
                if self.fixed_trial_step is None:
                    lam2 = lam
                else:
                    lam2 = self.fixed_trial_step
                R_G *= lam + lam2
                axpy(lam * lam2, dR_G, R_G)
                
            self.timer.start('precondition')
            psit_xG[:] += self.preconditioner(R_xG, kpt, ekin_x)
            self.timer.stop('precondition')
            
        self.timer.stop('RMM-DIIS')
        error = comm.sum(error)
        return error, psit_nG



from math import pi

import numpy as np
from ase.units import Hartree

from gpaw.xc import XC
from gpaw.xc.sic import SIC
from gpaw.atom.generator import Generator
from gpaw.atom.configurations import parameters
from gpaw.utilities import hartree


class NSCFSIC:
    def __init__(self, paw):
        self.paw = paw

    def calculate(self):
        ESIC = 0
        xc = self.paw.hamiltonian.xc
        assert xc.type == 'LDA'

        # Calculate the contribution from the core orbitals
        for a in self.paw.density.D_asp:
            setup = self.paw.density.setups[a]
            # TODO: Use XC which has been used to calculate the actual
            # calculation.
            # TODO: Loop over setups, not atoms.
            print('Atom core SIC for ', setup.symbol)
            print('%10s%10s%10s' % ('E_xc[n_i]', 'E_Ha[n_i]', 'E_SIC'))
            g = Generator(setup.symbol, xcname='LDA', nofiles=True, txt=None)
            g.run(**parameters[setup.symbol])
            njcore = g.njcore
            for f, l, e, u in zip(g.f_j[:njcore], g.l_j[:njcore],
                                  g.e_j[:njcore], g.u_j[:njcore]):
                # Calculate orbital density
                # NOTE: It's spherically symmetrized!
                #n = np.dot(self.f_j,
                assert l == 0, ('Not tested for l>0 core states')
                na = np.where(abs(u) < 1e-160, 0,u)**2 / (4 * pi)
                na[1:] /= g.r[1:]**2
                na[0] = na[1]
                nb = np.zeros(g.N)
                v_sg = np.zeros((2, g.N))
                e_g = np.zeros(g.N)
                vHr = np.zeros(g.N)
                Exc = xc.calculate_spherical(g.rgd, np.array([na, nb]), v_sg)
                hartree(0, na * g.r * g.dr, g.r, vHr)
                EHa = 2*pi*np.dot(vHr*na*g.r , g.dr)
                print(('%10.2f%10.2f%10.2f' % (Exc * Hartree, EHa * Hartree,
                                               -f*(EHa+Exc) * Hartree)))
                ESIC += -f*(EHa+Exc)
                
        sic = SIC(finegrid=True, coulomb_factor=1, xc_factor=1)
        sic.initialize(self.paw.density, self.paw.hamiltonian, self.paw.wfs)
        sic.set_positions(self.paw.atoms.get_scaled_positions())
        
        print('Valence electron sic ')
        print('%10s%10s%10s%10s%10s%10s' % ('spin', 'k-point', 'band',
                                            'E_xc[n_i]', 'E_Ha[n_i]', 'E_SIC'))
        assert len(self.paw.wfs.kpt_u)==1, ('Not tested for bulk calculations')
        
        for s, spin in sic.spin_s.items():
            spin.initialize_orbitals()
            spin.update_optimal_states()
            spin.update_potentials()

            n = 0
            for xc, c in zip(spin.exc_m, spin.ecoulomb_m):
                print(('%10i%10i%10i%10.2f%10.2f%10.2f' %
                       (s, 0, n, -xc * Hartree, -c * Hartree,
                        2 * (xc + c) * Hartree)))
                n += 1

            ESIC += spin.esic
            
        print('Total correction for self-interaction energy:')
        print('%10.2f eV' % (ESIC * Hartree))
        print('New total energy:')
        total = (ESIC * Hartree + self.paw.get_potential_energy() +
                 self.paw.get_reference_energy())
        print('%10.2f eV' % total)
        return total



# Copyright (C) 2010  CAMd
# Please see the accompanying LICENSE file for further information.

"""This module provides all the classes and functions associated with the
evaluation of exact exchange with k-point sampling."""

from __future__ import print_function
from math import pi, sqrt
from time import time

import numpy as np

from gpaw.utilities import pack, unpack2, packed_index, logfile
from gpaw.lfc import LFC
from gpaw.wavefunctions.pw import PWDescriptor, PWWaveFunctions
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.xc.hybrid import HybridXCBase


class KPoint:
    def __init__(self, kd, kpt=None):
        """Helper class for parallelizing over k-points.

        Placeholder for wave functions, occupation numbers,
        projections, and global k-point index."""
        
        self.kd = kd
        
        if kpt is not None:
            self.psit_nG = kpt.psit_nG
            self.f_n = kpt.f_n / kpt.weight / kd.nbzkpts * 2 / kd.nspins
            self.eps_n = kpt.eps_n
            self.weight = np.array([kpt.weight])

            self.P_ani = kpt.P_ani
            self.k = kpt.k
            self.s = kpt.s
 
        self.requests = []
        
    def next(self):
        """Create empty object.

        Data will be received from other processor."""
        
        kpt = KPoint(self.kd)

        # intialize array for receiving:
        kpt.psit_nG = np.empty_like(self.psit_nG)
        kpt.f_n = np.empty_like(self.f_n)
        kpt.eps_n = np.empty_like(self.eps_n)
        kpt.weight = np.empty_like(self.weight)

        # Total number of projector functions:
        I = sum([P_ni.shape[1] for P_ni in self.P_ani.values()])
        
        kpt.P_nI = np.empty((len(kpt.f_n), I), complex)

        kpt.P_ani = {}
        I1 = 0
        for a, P_ni in self.P_ani.items():
            I2 = I1 + P_ni.shape[1]
            kpt.P_ani[a] = kpt.P_nI[:, I1:I2]
            I1 = I2

        kpt.k = (self.k + 1) % self.kd.nibzkpts
        kpt.s = self.s
        
        return kpt
        
    def start_sending(self, rank):
        P_nI = np.hstack([P_ni for P_ni in self.P_ani.values()])
        P_nI = np.ascontiguousarray(P_nI)
        self.requests += [
            self.kd.comm.send(self.psit_nG, rank, block=False, tag=1),
            self.kd.comm.send(self.f_n, rank, block=False, tag=2),
            self.kd.comm.send(self.eps_n, rank, block=False, tag=3),
            self.kd.comm.send(self.weight, rank, block=False, tag=4),
            self.kd.comm.send(P_nI, rank, block=False, tag=5)]
        
    def start_receiving(self, rank):
        self.requests += [
            self.kd.comm.receive(self.psit_nG, rank, block=False, tag=1),
            self.kd.comm.receive(self.f_n, rank, block=False, tag=2),
            self.kd.comm.receive(self.eps_n, rank, block=False, tag=3),
            self.kd.comm.receive(self.weight, rank, block=False, tag=4),
            self.kd.comm.receive(self.P_nI, rank, block=False, tag=5)]

    def wait(self):
        self.kd.comm.waitall(self.requests)
        self.requests = []
        

class HybridXC(HybridXCBase):
    orbital_dependent = True
    def __init__(self, name, hybrid=None, xc=None, gygi=False,
                 alpha=None, skip_gamma=False, ecut=None,
                 etotflag = False, acdf=False, coredensity=True,
                 logfilename='-', bands=None, core_valence=True):
        """Mix standard functionals with exact exchange.

        bands: list or None
            List of bands to calculate energy for.  Default is None
            meaning do all bands.
        """

        self.alpha = alpha
        self.skip_gamma = skip_gamma
        self.gygi = gygi
        
        self.exx = 0.0
        self.etotflag = etotflag
        self.ecut = ecut
        self.fd = logfilename
        self.write_timing_information = True
        self.bands = bands
        self.acdf = acdf  # adiabatic-connection dissipation fluctuation for RPA correlation energy
        self.coredensity = coredensity
        self.core_valence = core_valence
        if self.acdf:
            self.exxacdf = 0.0
            self.etotflag = True
            print('etotflag is True')

        HybridXCBase.__init__(self, name, hybrid, xc)

    def log(self, *args, **kwargs):
        print(file=self.fd, *args, **kwargs)
        self.fd.flush()

    def calculate_radial(self, rgd, n_sLg, Y_L, v_sg,
                         dndr_sLg=None, rnablaY_Lv=None,
                         tau_sg=None, dedtau_sg=None):
        return self.xc.calculate_radial(rgd, n_sLg, Y_L, v_sg,
                                        dndr_sLg, rnablaY_Lv)
    
    def calculate_paw_correction(self, setup, D_sp, dEdD_sp=None,
                                 addcoredensity=True, a=None):
        addcoredensity = self.coredensity # XXX overwrites input

        return self.xc.calculate_paw_correction(setup, D_sp, dEdD_sp,
                                 addcoredensity, a)
    
    def initialize(self, density, hamiltonian, wfs, occupations):
        self.xc.initialize(density, hamiltonian, wfs, occupations)
        self.nspins = wfs.nspins
        self.setups = wfs.setups
        self.density = density
        self.kpt_u = wfs.kpt_u
        
        self.gd = density.gd
        self.kd = wfs.kd
        self.bd = wfs.bd
        if self.bd.comm.size > 1:
            raise ValueError('Band parallelization not supported by hybridk')
        self.wfs = wfs

        self.world = wfs.world

        self.fd = logfile(self.fd, self.world.rank)

        N = self.gd.N_c.prod()
        vol = self.gd.dv * N
        
        if self.alpha is None:
            # XXX ?
            self.alpha = 6 * vol**(2 / 3.0) / pi**2
            
        if self.ecut is None:
            self.ecut = 0.5 * pi**2 / (self.gd.h_cv**2).sum(1).max() * 0.9999
            
        self.bzq_qc = self.kd.get_bz_q_points()
        qd = KPointDescriptor(self.bzq_qc)
        q0 = self.kd.where_is_q(np.zeros(3), self.bzq_qc)
        
        self.pwd = PWDescriptor(self.ecut, self.gd, complex, kd=qd)

        G2_qG = self.pwd.G2_qG
        G2_qG[q0][0] = 117.0
        self.iG2_qG = [1.0 / G2_G for G2_G in G2_qG]
        G2_qG[q0][0] = 0.0
        self.iG2_qG[q0][0] = 0.0

        self.gamma = (vol / (2 * pi)**2 * sqrt(pi / self.alpha) *
                      self.kd.nbzkpts)

        for q in range(self.kd.nbzkpts):
            self.gamma -= np.dot(np.exp(-self.alpha * G2_qG[q]),
                                 self.iG2_qG[q])

        self.iG2_qG[q0][0] = self.gamma
        
        self.ghat = LFC(self.gd,
                        [setup.ghat_l for setup in density.setups],
                        qd, dtype=complex)
        
        self.log('Value of alpha parameter:', self.alpha)
        self.log('Value of gamma parameter:', self.gamma)
        self.log('Cutoff energy:', self.ecut, 'Hartree')
        self.log('%d x %d x %d k-points' % tuple(self.kd.N_c))

    def set_positions(self, spos_ac):
        self.ghat.set_positions(spos_ac)
        self.spos_ac = spos_ac

    def calculate(self, gd, n_sg, v_sg=None, e_g=None):
        # Normal XC contribution:
        exc = self.xc.calculate(gd, n_sg, v_sg, e_g)

         # Add EXX contribution:
        return exc + self.exx

    def calculate_exx(self):
        """Non-selfconsistent calculation."""

        kd = self.kd
        K = kd.nibzkpts
        W = self.world.size // self.nspins
        parallel = (W > 1)

        self.log("%d CPU's used for %d IBZ k-points" % (W, K))
        self.log('Spins:', self.nspins)

        if self.etotflag and not self.gygi:
            self.nbandstmp = 0
            for s in range(self.nspins):
                kpt1_k = [KPoint(kd, kpt)
                          for kpt in self.kpt_u if kpt.s == s]
                for kpt1 in kpt1_k:
                    for n1 in range(self.bd.nbands):
                        f_n = kpt1.f_n[n1]
                        if np.abs(f_n) < 1e-10:
                            self.nbandstmp = max(self.nbandstmp, n1)
                            break
                    else:
                        self.nbandstmp = self.bd.nbands

            tmp = np.zeros(kd.comm.size, dtype=int)
            kd.comm.all_gather(np.array([self.nbandstmp]), tmp)
            self.nbands = tmp.max()
        else:
            self.nbands = self.bd.nbands
        
        B = self.nbands
        self.log('Number of bands calculated:', B)
        self.log('Number of valence electrons:', self.setups.nvalence)

        E = B - self.setups.nvalence / 2.0  # empty bands
        self.npairs = (K * kd.nbzkpts - 0.5 * K**2) * (B**2 - E**2)
        self.log('Approximate number of pairs:', self.npairs)

        if not self.etotflag:
            self.exx_skn = np.zeros((self.nspins, K, B))
            self.debug_skn = np.zeros((self.nspins, K, B))

        for s in range(self.nspins):
            kpt1_q = [KPoint(kd, kpt)
                      for kpt in self.kpt_u if kpt.s == s]
            kpt2_q = kpt1_q[:]

            if len(kpt1_q) == 0:
                # No s-spins on this CPU:
                continue

            # Send rank:
            srank = kd.get_rank_and_index(s, (kpt1_q[0].k - 1) % K)[0]
            # Receive rank:
            rrank = kd.get_rank_and_index(s, (kpt1_q[-1].k + 1) % K)[0]

            # Shift k-points K - 1 times:
            for i in range(K):
                if i < K - 1:
                    if parallel:
                        kpt = kpt2_q[-1].next()
                        kpt.start_receiving(rrank)
                        kpt2_q[0].start_sending(srank)
                    else:
                        kpt = kpt2_q[0]

                for kpt1, kpt2 in zip(kpt1_q, kpt2_q):
                    for k, ik in enumerate(kd.bz2ibz_k):
                        if ik == kpt2.k:
                            self.apply(kpt1, kpt2, k)

                if i < K - 1:
                    if parallel:
                        kpt.wait()
                        kpt2_q[0].wait()
                    kpt2_q.pop(0)
                    kpt2_q.append(kpt)

        if self.etotflag:
            if self.acdf:
                self.exxacdf = self.world.sum(self.exxacdf[0])
                self.exx = self.exxacdf
            else:
                self.exx = self.world.sum(self.exx)
            self.exx += self.calculate_exx_paw_correction()
                
        else:
            for kpt in self.kpt_u:
                for a, D_sp in self.density.D_asp.items():
                    setup = self.setups[a]
                    for D_p in D_sp:
                        D_ii = unpack2(D_p)
                        ni = len(D_ii)
                        P_ni = kpt.P_ani[a]
                        for i1 in range(ni):
                            for i2 in range(ni):
                                A = 0.0
                                for i3 in range(ni):
                                    p13 = packed_index(i1, i3, ni)
                                    for i4 in range(ni):
                                        p24 = packed_index(i2, i4, ni)
                                        A += setup.M_pp[p13, p24] * D_ii[i3, i4]
                                self.exx_skn[kpt.s, kpt.k] -= \
                                    (self.hybrid * A *
                                     P_ni[:, i1].conj() * P_ni[:, i2]).real
                                
                                p12 = packed_index(i1, i2, ni)
                                if self.core_valence:
                                    if setup.X_p is not None:
                                        self.exx_skn[kpt.s, kpt.k] -= self.hybrid * \
                                                                      (P_ni[:, i1].conj() * setup.X_p[p12] *
                                                                       P_ni[:, i2]).real / self.nspins

            
            self.world.sum(self.exx_skn)
            self.exx = 0.0
            for kpt in self.kpt_u:
                self.exx += 0.5 * np.dot(kpt.f_n, self.exx_skn[kpt.s, kpt.k])
            self.exx = self.world.sum(self.exx)

            for a, D_sp in self.density.D_asp.items():
                setup = self.setups[a]
                if self.coredensity:
                    self.exx += self.hybrid * setup.ExxC
                if self.core_valence:
                    self.exx -= self.hybrid * 0.5 * np.dot(D_sp.sum(0), setup.X_p)

            self.world.sum(self.debug_skn)
            assert (self.debug_skn == self.kd.nbzkpts * B).all()
    
    def apply(self, kpt1, kpt2, k):
        k1_c = self.kd.ibzk_kc[kpt1.k]
        k20_c = self.kd.ibzk_kc[kpt2.k]
        k2_c = self.kd.bzk_kc[k]
        q_c = k2_c - k1_c
        N_c = self.gd.N_c

        q = self.kd.where_is_q(q_c, self.bzq_qc)
        
        q_c = self.bzq_qc[q]
        eik1r_R = np.exp(2j * pi * np.dot(np.indices(N_c).T, k1_c / N_c).T)
        eik2r_R = np.exp(2j * pi * np.dot(np.indices(N_c).T, k20_c / N_c).T)
        eiqr_R = np.exp(2j * pi * np.dot(np.indices(N_c).T, q_c / N_c).T)

        same = abs(k1_c - k2_c).max() < 1e-9

        iG2_G = self.iG2_qG[q]
            
        N = N_c.prod()
        vol = self.gd.dv * N
        nspins = self.nspins

        fcut = 1e-10
        is_ibz2 = abs(k2_c - self.kd.ibzk_kc[kpt2.k]).max() < 1e-9
        
        for n1 in range(self.nbands):
            f1 = kpt1.f_n[n1]
            e1 = kpt1.eps_n[n1]
            for n2 in range(self.nbands):
                if same:
                    assert is_ibz2
                    if n2 > n1:
                        continue
                elif is_ibz2:
                    if kpt1.k > kpt2.k:
                        if n2 > n1:
                            continue
                    else:
                        if n2 >= n1:
                            continue
                        
                f2 = kpt2.f_n[n2]
                e2 = kpt2.eps_n[n2]

                x = 1.0
                if same and n1 == n2:
                    x = 0.5

                if not self.etotflag:
                    self.debug_skn[kpt1.s, kpt1.k, n1] += x
                    if is_ibz2:
                        self.debug_skn[kpt2.s, kpt2.k, n2] += x

                if self.etotflag and not self.gygi:
                    if abs(f1) < fcut or abs(f2) < fcut:
                        continue
                else:
                    if abs(f1) < fcut and abs(f2) < fcut:
                        continue

                if self.bands is not None:
                    if not (n1 in self.bands or is_ibz2 and n2 in self.bands):
                        continue

                if self.skip_gamma and same:
                    continue
                
                t0 = time()
                nt_R = self.calculate_pair_density(n1, n2, kpt1, kpt2, q, k,
                                                   eik1r_R, eik2r_R, eiqr_R,
                                                   is_ibz2)
                nt_G = self.pwd.fft(nt_R, q) / N
                vt_G = nt_G.copy()
                vt_G *= -pi * vol * iG2_G
                e = np.vdot(nt_G, vt_G).real * nspins * self.hybrid * x
                
                if self.etotflag:
                    if self.acdf:
                        if self.gygi and same:
                            self.exxacdf += f2 * e * kpt1.weight
                        else:
                            self.exxacdf += 0.5 * (f1 * (1-np.sign(e2-e1)) * e +
                                                   f2 * (1-np.sign(e1-e2)) * e ) * kpt1.weight
                    else:
                        self.exx += f2 * e * kpt1.weight[0] * f1 * self.kd.nbzkpts * nspins / 2
                else:
                    self.exx_skn[kpt1.s, kpt1.k, n1] += 2 * f2 * e

                if is_ibz2:
                    if self.etotflag:
                        if self.acdf:
                            if self.gygi and same:
                                self.exxacdf += f1 * e * kpt2.weight
                            else:
                                self.exxacdf += 0.5 * (f1 * (1-np.sign(e2-e1)) * e +
                                                       f2 * (1-np.sign(e1-e2)) * e ) * kpt2.weight
                        else:
                            self.exx += f1 * e * kpt2.weight[0] * f2 * self.kd.nbzkpts * nspins / 2
                    else:
                        self.exx_skn[kpt2.s, kpt2.k, n2] += 2 * f1 * e
                    
                if self.write_timing_information:
                    t = time() - t0
                    self.log('Time for first pair-density:', t, 'seconds')
                    self.log('Estimated total time',
                             t * self.npairs / self.world.size, 'seconds')
                    self.write_timing_information = False

    def calculate_exx_paw_correction(self):
        exx = 0
        deg = 2 // self.nspins  # spin degeneracy
        for a, D_sp in self.density.D_asp.items():
            setup = self.setups[a]
            for D_p in D_sp:
                D_ii = unpack2(D_p)
                ni = len(D_ii)

                for i1 in range(ni):
                    for i2 in range(ni):
                        A = 0.0
                        for i3 in range(ni):
                            p13 = packed_index(i1, i3, ni)
                            for i4 in range(ni):
                                p24 = packed_index(i2, i4, ni)
                                A += setup.M_pp[p13, p24] * D_ii[i3, i4]
                        exx -= self.hybrid / deg * D_ii[i1, i2] * A

                if self.core_valence:
                    if setup.X_p is not None:
                        exx -= self.hybrid * np.dot(D_p, setup.X_p)
            if self.coredensity:
                exx += self.hybrid * setup.ExxC
        return exx

    def calculate_pair_density(self, n1, n2, kpt1, kpt2, q, k,
                               eik1r_R, eik2r_R, eiqr_R, ibz2):
        if isinstance(self.wfs, PWWaveFunctions):
            psit1_R = self.wfs.pd.ifft(kpt1.psit_nG[n1]) * eik1r_R
            psit2_R = self.wfs.pd.ifft(kpt2.psit_nG[n2]) * eik2r_R
        else:
            psit1_R = kpt1.psit_nG[n1]
            psit2_R = kpt2.psit_nG[n2]

        if ibz2:
            psit2_R = psit2_R
        else:
            psit2_R = np.asarray(self.kd.transform_wave_function(psit2_R, k),
                                 complex)
        nt_R = psit1_R.conj() * psit2_R

        s = self.kd.sym_k[k]
        time_reversal = self.kd.time_reversal_k[k]
        k2_c = self.kd.ibzk_kc[kpt2.k]

        Q_aL = {}
        for a, P1_ni in kpt1.P_ani.items():
            P1_i = P1_ni[n1]

            b = self.kd.symmetry.a_sa[s, a]
            S_c = (np.dot(self.spos_ac[a], self.kd.symmetry.op_scc[s]) -
                   self.spos_ac[b])
            assert abs(S_c.round() - S_c).max() < 1e-13
            x = np.exp(2j * pi * np.dot(k2_c, S_c))
            P2_i = np.dot(self.setups[a].R_sii[s], kpt2.P_ani[b][n2]) * x
            if time_reversal:
                P2_i = P2_i.conj()

            if ibz2:
                P2_i = kpt2.P_ani[a][n2]

            D_ii = np.outer(P1_i.conj(), P2_i)
            D_p = pack(D_ii)
            Q_aL[a] = np.dot(D_p, self.setups[a].Delta_pL)

        self.ghat.add(nt_R, Q_aL, q)
        return nt_R / eiqr_R



from __future__ import print_function
from gpaw.tddft import photoabsorption_spectrum
from ase import Atoms
from gpaw import GPAW
from gpaw.lcaotddft import LCAOTDDFT
import gpaw.lrtddft as lrtddft
from gpaw.mpi import world

N = 2
xc = 'oldLDA'
c = 0
h = 0.4
b = 'sz(dzp)'
sy = 'H'+str(N)
positions = []
for i in range(N):
    positions.append([0.00,0.00,i*0.7])
atoms = Atoms(symbols=sy, positions = positions)
atoms.center(vacuum=3)
print(atoms)
# LCAO-RT-TDDFT
parallel = {}
if world.size > 3:
    parallel['band'] = 2

calc = LCAOTDDFT(mode='lcao', xc=xc, h=h, basis=b, nbands=N,
                 dtype=complex, charge=c, convergence={'density':1e-6}, 
                 propagator='cn',
                 parallel=parallel)
atoms.set_calculator(calc)
atoms.get_potential_energy()
dmfile = sy+'_lcao_'+b+'_rt_z.dm'+str(world.size)
specfile = sy+'_lcao_'+b+'_rt_z.spectrum'+str(world.size)
calc.absorption_kick([0.0,0,0.001])
calc.propagate(10, 20, dmfile)
if world.rank == 0:
    photoabsorption_spectrum(dmfile, specfile)

if 0:
    # Reference RS-LR-TDDFT
    calc = GPAW(xc=xc, h=h, charge=c, width=0, nbands=4)
    atoms.set_calculator(calc)
    atoms.get_potential_energy()
    lr = lrtddft.LrTDDFT(calc, finegrid=0)
    lr.diagonalize()
    if world.rank == 0:
        lrtddft.photoabsorption_spectrum(lr, sy+'_rs_lr.spectrum', e_min=0.0, e_max=40)

    # Reference LCAO-LR-TDDFT
    calc = GPAW(mode='lcao', xc=xc, h=h, basis=b, charge=c, width=0)
    atoms.set_calculator(calc)
    atoms.get_potential_energy()
    lr = lrtddft.LrTDDFT(calc, finegrid=0)
    lr.diagonalize()
    if world.rank == 0:
        lrtddft.photoabsorption_spectrum(lr, sy+'_lcao_'+b+'_lr.spectrum', e_min=0.0, e_max=400)

    # XXX Actually check that the spectra match



from ase import Atoms, io
from ase.calculators.vdwcorrection import vdWTkatchenko09prl
from ase.structure import molecule
from ase.parallel import barrier

from gpaw import GPAW
from gpaw.cluster import Cluster
from gpaw.analyse.hirshfeld import HirshfeldDensity, HirshfeldPartitioning
from gpaw.analyse.vdwradii import vdWradii
from gpaw.test import equal

h = 0.4
s = Cluster(molecule('Na2'))
s.minimal_box(3., h=h)

out_traj = 'Na2.traj'
out_txt = 'Na2.txt'
                       
cc = GPAW(h=h, xc='PBE', txt=out_txt)

# this is needed to initialize txt output
cc.initialize(s)

c = vdWTkatchenko09prl(HirshfeldPartitioning(cc),
                       vdWradii(s.get_chemical_symbols(), 'PBE'))
s.set_calculator(c)
E = s.get_potential_energy()
F_ac = s.get_forces()
s.write(out_traj)

barrier()

# test I/O, accuracy due to text output
accuracy = 1.e-5
for fname in [out_traj, out_txt]:
    s_out = io.read(fname)
    ##print s_out.get_potential_energy(), E
    ##print s_out.get_forces()
    equal(s_out.get_potential_energy(), E, accuracy)
    for fi, fo in zip(F_ac, s_out.get_forces()):
        equal(fi, fo, accuracy)



from __future__ import print_function
import os
from gpaw import GPAW, restart
from ase import Atoms
from gpaw.test import equal
from math import sqrt
import numpy as np

modes = ['gpw']
try:
    import _gpaw_hdf5
    modes.append('hdf5')
except ImportError:
    pass

d = 3.0
atoms = Atoms('Na3', positions=[( 0, 0, 0),
                              ( 0, 0, d),
                              ( 0, d*sqrt(3./4.), d/2.)],
                   magmoms=[1.0, 1.0, 1.0],
                   cell=(3.5, 3.5, 4.+2/3.),
                   pbc=True)

# Only a short, non-converged calcuation
conv = {'eigenstates': 1.24, 'energy':2e-1, 'density':1e-1}
calc = GPAW(h=0.30, nbands=3,
            setups={'Na': '1'},
            convergence=conv)
atoms.set_calculator(calc)
e0 = atoms.get_potential_energy()
niter0 = calc.get_number_of_iterations()
f0 = atoms.get_forces()
m0 = atoms.get_magnetic_moments()
eig00 = calc.get_eigenvalues(spin=0)
eig01 = calc.get_eigenvalues(spin=1)
# Write the restart file(s)
for mode in modes:
    calc.write('tmp.%s' % mode)

del atoms, calc
# Try restarting from all the files
for mode in modes:
    atoms, calc = restart('tmp.%s' % mode)
    e1 = atoms.get_potential_energy()
    try: # number of iterations needed in restart
        niter1 = calc.get_number_of_iterations()
    except: pass
    f1 = atoms.get_forces()
    m1 = atoms.get_magnetic_moments()
    eig10 = calc.get_eigenvalues(spin=0)
    eig11 = calc.get_eigenvalues(spin=1)
    print(e0, e1)
    equal(e0, e1, 1e-10)
    print(f0, f1)
    for ff0, ff1 in zip(f0, f1):
        err = np.linalg.norm(ff0-ff1)
        assert err <= 1e-10
    print(m0, m1)
    for mm0, mm1 in zip(m0, m1):
        equal(mm0, mm1, 1e-10)
    print('A',eig00, eig10)
    for eig0, eig1 in zip(eig00, eig10):
        equal(eig0, eig1, 1e-10)
    print('B',eig01, eig11)
    for eig0, eig1 in zip(eig01, eig11):
        equal(eig0, eig1, 1e-10)

    # Check that after restart everything is writable
    calc.write('tmp2.%s' % mode)



import numpy as np
from ase.units import Bohr
from ase.transport.tools import dagger

from gpaw.transport.selfenergy import LeadSelfEnergy
from gpaw.transport.greenfunction import GreenFunction
#from ase.transport.selfenergy import LeadSelfEnergy
#from ase.transport.greenfunction import GreenFunction

from gpaw.fd_operators import Laplace
from gpaw.grid_descriptor import GridDescriptor
from gpaw.lfc import NewLocalizedFunctionsCollection as LFC
import gpaw.mpi as mpi
from gpaw.utilities.tools import tri2full

''' differ from version 4 by selected scaning range in z direction by values in Angstrom '''

class LocalizedFunctions:

    def __init__(self, gd, f_iG, corner_c, index=None, vt_G=None):
        self.gd = gd
        #assert not gd.is_non_orthogonal()
        self.size_c = np.array(f_iG.shape[1:4])
        self.f_iG = f_iG
        self.corner_c = corner_c
        self.index = index
        self.vt_G = vt_G
   
    def periodic(self):
        trans_c = [np.array([0, 0, 0]), np.array([0, 1, 0]), np.array([-1,  0,  0]), np.array([ 0, -1,  0]), 
                   np.array([1, 1, 0]), np.array([ 1, -1,  0]), np.array([-1,  1,  0]), np.array([-1, -1,  0]), 
                   np.array([ 1,  0,  0])]
        trans_c[:] *= np.array([self.gd.N_c[0],self.gd.N_c[1],0])
        
        self.periodic_list = trans_c+self.corner_c
 
        list = []
        for corner in self.periodic_list:
            list.append(LocalizedFunctions(self.gd,self.f_iG,
                                        corner_c=corner,
                                        index=self.index,
                                        vt_G=self.vt_G))
        return list



    def __len__(self):
        return len(self.f_iG)

    def apply_t(self):
        """Apply kinetic energy operator and return new object."""
        p = 2  # padding
        newsize_c = self.size_c + 2 * p
        gd = GridDescriptor(N_c=newsize_c + 1,
                            cell_cv=self.gd.h_c * (newsize_c + 1),
                            pbc_c=False,
                            comm=mpi.serial_comm)
        T = Laplace(gd, scale =1/2., n=p)
        f_ig = np.zeros((len(self.f_iG),) + tuple(newsize_c))
        f_ig[:, p:-p, p:-p, p:-p] = self.f_iG
        Tf_iG = np.empty_like(f_ig)
        T.apply(f_ig, Tf_iG)
        return LocalizedFunctions(self.gd, Tf_iG, self.corner_c - p,
                                  self.index)
        
    def overlap(self, other):
        start_c = np.maximum(self.corner_c, other.corner_c)
        stop_c = np.minimum(self.corner_c + self.size_c,
                            other.corner_c + other.size_c)
        if (start_c < stop_c).all():
            astart_c = start_c - self.corner_c
            astop_c = stop_c - self.corner_c
            a_iG = self.f_iG[:,
                astart_c[0]:astop_c[0],
                astart_c[1]:astop_c[1],
                astart_c[2]:astop_c[2]].copy().reshape((len(self.f_iG), -1))
            bstart_c = start_c - other.corner_c
            bstop_c = stop_c - other.corner_c
            b_iG = other.f_iG[:,
                bstart_c[0]:bstop_c[0],
                bstart_c[1]:bstop_c[1],
                bstart_c[2]:bstop_c[2]].copy().reshape((len(other.f_iG), -1))
            if self.vt_G is not None:
                x = self.vt_G[start_c[0]:stop_c[0],
                              start_c[1]:stop_c[1],
                              start_c[2]:stop_c[2]].copy().reshape((-1,))
                #print 'start_c:',start_c
                #print 'stop_c:',stop_c
                #print 'shape of x', np.shape(x)
                #print 'shape of a_iG', np.shape(a_iG)
                a_iG *= x    
     
            return self.gd.dv * np.inner(a_iG, b_iG)
        else:
            return None
    

    def __or__(self, other):
        if isinstance(other, LocalizedFunctions):
            return self.overlap(other)

        # other is a potential:
        vt_G = other
        return LocalizedFunctions(self.gd, self.f_iG, self.corner_c,
                                  self.index, vt_G)

class WannierFunction(LocalizedFunctions):
    def __init__(self, gd, wanf_G, corner_c, index=None):
        LocalizedFunctions.__init__(self, gd, wanf_G[np.newaxis, :, :, :],
                                    corner_c, index)

class AtomCenteredFunctions(LocalizedFunctions):
    def __init__(self, gd, spline_j, spos_c, index=None):
        rcut = max([spline.get_cutoff() for spline in spline_j])
        corner_c = np.ceil(spos_c * gd.N_c - rcut / gd.h_c).astype(int)
        size_c = np.ceil(spos_c * gd.N_c + rcut / gd.h_c).astype(int) - corner_c
        smallgd = GridDescriptor(N_c=size_c + 1,
                                 cell_cv=gd.h_c * (size_c + 1),
                                 pbc_c=False,
                                 comm=mpi.serial_comm)
        lfc = LFC(smallgd, [spline_j])
        lfc.set_positions((spos_c[np.newaxis, :] * gd.N_c - corner_c + 1) /
                          smallgd.N_c)
        ni = lfc.Mmax
        f_iG = smallgd.zeros(ni)
        lfc.add(f_iG, {0: np.eye(ni)})
        LocalizedFunctions.__init__(self, gd, f_iG, corner_c, 
                                     index=index)


class STM:
    def __init__(self, srf, tip, srf_pl, tip_pl, eta1=1e-4, eta2=1e-4):
        self.tip = tip                     # tip calcualtor
        self.tip_pl = tip_pl               # tip periodic layers calculator
        self.srf = srf                     # surface calculator
        self.srf_pl = srf_pl               # surface periodic layers calculator
        self.eta1 = eta1
        self.eta2 = eta2 
        self.tgd = tip.gd                  
        self.sgd = srf.gd 
        
        #assert tgd.h_cv == sgd.h_cv
        #assert not (tgd.h_c - sgd.h_c).any()

    def initialize(self, srf_atoms, tip_atoms, srf_coupling_atoms, tip_coupling_atoms, 
                   srf_pl_atoms, tip_pl_atoms, energies, bias=0):
        """
        srf_atoms: the atom index of onsite surface atoms 
        tip_atoms: the atom index of onsite tip atoms
        energies: list of energies, for which the transmission function should be evaluated
        bias: will be used to precalculate the greenfunctions of surface and tip
        """
        self.bias = bias
        self.energies = energies
        self.tip_atoms = tip_atoms
        self.srf_atoms = srf_atoms
        self.srf_coupling_atoms = srf_coupling_atoms
        self.tip_coupling_atoms = tip_coupling_atoms
        #align potentials for onsite surface, onsite tip and tip principle layers according to surface_pl
        srf_alignv = self.align_v(calc1 = self.srf_pl, index1=srf_pl_atoms[0], calc2=self.srf, index2=srf_atoms[0])         
        tip_alignv = srf_alignv + self.tip.get_fermi_level() - self.srf.get_fermi_level() #- bias
        tip_pl_alignv = self.align_v(calc1=self.tip, index1=tip_atoms[-1], calc2=self.tip_pl, index2=tip_pl_atoms[-1])
        tip_pl_alignv += tip_alignv
        self.srf_alignv = srf_alignv
        self.tip_alignv = tip_alignv

        #hamiltoian and overlap matrix of onsite tip and srf
        self.h1, self.s1 = Lead_HS(calc = self.tip, atom_list = tip_atoms, shift = srf_alignv-bias).get_HS()
        print('shape of onsite tip', np.shape(self.h1), np.shape(self.s1))
        self.h2, self.s2 = Lead_HS(calc = self.srf, atom_list = srf_atoms, shift = tip_alignv).get_HS()        
        print('shape of onsite srf', np.shape(self.h2), np.shape(self.s2))

        #hamiltoian and overlap matrix of two tip and srf principle layers
        self.h10, self.s10 = Lead_HS(calc = self.tip_pl, atom_list = tip_pl_atoms, shift = tip_pl_alignv).get_HS() #2 and only 2 principle layers
        print('shape of pl tip', np.shape(self.h10), np.shape(self.s10)) 
        self.h20, self.s20 = Lead_HS(calc = self.srf_pl, atom_list = srf_pl_atoms, shift = -bias).get_HS() #2 and only 2 principle layers
        print('shape of pl srf', np.shape(self.h20), np.shape(self.s20))

        nbf1, nbf2 = len(self.h1), len(self.h2)       #No. of basis functions of onsite tip and surface atoms
        pl1, pl2 = len(self.h10)/2, len(self.h20)/2   #No. of basis functions per principle layer of tip and srf
        nenergies = len(energies)        

        #periodic part of the tip
        hs1_dii = self.h10[:pl1, :pl1], self.s10[:pl1, :pl1]
        hs1_dij = self.h10[:pl1, pl1:2*pl1], self.s10[:pl1, pl1:2*pl1]
        #coupling between per. and non. per part of the tip
        h1_im = np.zeros((pl1, nbf1), complex)
        s1_im = np.zeros((pl1, nbf1), complex)
        h1_im[:pl1, :pl1], s1_im[:pl1, :pl1] = hs1_dij
        hs1_dim = [h1_im, s1_im]

        #periodic part the surface 
        hs2_dii = self.h20[:pl2, :pl2], self.s20[:pl2, :pl2]
        hs2_dij = self.h20[pl2:2*pl2, :pl2], self.s20[pl2:2*pl2, :pl2]
        #coupling between per. and non. per part of the surface
        h2_im = np.zeros((pl2, nbf2), complex)
        s2_im = np.zeros((pl2, nbf2), complex)
        h2_im[-pl2:, -pl2:], s2_im[-pl2:, -pl2:] = hs2_dij
        hs2_dim = [h2_im, s2_im]
        #tip and surface greenfunction 
        self.selfenergy1 = LeadSelfEnergy(hs1_dii, hs1_dij, hs1_dim)
        self.selfenergy2 = LeadSelfEnergy(hs2_dii, hs2_dij, hs2_dim)
        self.greenfunction1 = GreenFunction(self.h1, self.s1,[self.selfenergy1])
        self.greenfunction2 = GreenFunction(self.h2, self.s2,[self.selfenergy2])
        #print 'shape of greenfunction tip:', np.shape(self.greenfunction1)
        #print 'shape of greenfunction srf:', np.shape(self.greenfunction2)
        #shift the bands due to the bias
        self.selfenergy1.set_bias(0.0)
        self.selfenergy2.set_bias(bias)

        # extend the surface potentials and basis functions
        self.setup_stm()


        #tip and surface greenfunction matrices.
        nbf1_small = self.ni #XXX Change this for efficiency in the future
        nbf2_small = self.nj #XXX -||-
        coupling_list1 = range(nbf1-nbf1_small,nbf1)# XXX -||-
        coupling_list2 = range(0,nbf2_small)# XXX -||-
        self.gft1_emm = np.zeros((nenergies, nbf1_small, nbf1_small), complex)
        self.gft2_emm = np.zeros((nenergies, nbf2_small, nbf2_small), complex)

        for e, energy in enumerate(self.energies):
            gft1_mm = self.greenfunction1(energy)[coupling_list1]
            gft1_mm = np.take(gft1_mm, coupling_list1, axis=1)

            gft2_mm = self.greenfunction2(energy)[coupling_list2]
            gft2_mm = np.take(gft2_mm, coupling_list2, axis=1)

            self.gft1_emm[e] = gft1_mm
            self.gft2_emm[e] = gft2_mm

    def scan(self,height=None,zmin=None,zmax=None):
        if height is not None:
            self.current = np.zeros((self.srf.gd.N_c[0],self.srf.gd.N_c[1]))
            h = 0
            h = np.round(height/Bohr/self.srf.gd.h_c[2]).astype(int)
            for px in range(self.srf.gd.N_c[0]):
                for py in range(self.srf.gd.N_c[1]):
                    pos = np.array([px,py,h])
                    V_12 = self.get_Vts(pos)
                    I = self.get_current(V_12)
                    self.current[px,py] = I
                    print('I = ', I)
            return self.current
        if height is None:
            zmin_c = np.round(zmin/Bohr/self.srf.gd.h_c[2]).astype(int)
            zmax_c = np.round(zmax/Bohr/self.srf.gd.h_c[2]).astype(int)
            nz = zmax_c - zmin_c
            self.current = np.zeros((self.srf.gd.N_c[0],self.srf.gd.N_c[1],nz))
            for px in range(self.srf.gd.N_c[0]):
                for py in range(self.srf.gd.N_c[1]):
                    for pz in range(zmin_c,zmax_c):
                        # get coupleing between tip and surface Vts for a given tip position on original srf grids 
                        pos = np.array([px,py,pz])
                        V_12 = self.get_Vts(pos)
                        # get current values
                        I = self.get_current(V_12)
                        self.current[px,py,pz-zmin_c] = I
                        print(I)
            return self.current



    def get_current(self, v_12):
        # get transmission
        T_e = self.get_transmission(v_12)
        I = -np.trapz(x = self.energies, y = T_e)         
        return I



    def get_transmission(self, v_12):
        nenergies = len(self.energies)
        T_e = np.empty(nenergies,float)
        v_21 = dagger(v_12)
        for e, energy in enumerate(self.energies):
            gft1 = self.gft1_emm[e]
            gft2 = self.gft2_emm[e]            
            a1 = (gft1 - dagger(gft1))
            a2 = (gft2 - dagger(gft2))
            v12_a2 = np.dot(v_12, a2)
            v21_a1 = np.dot(v_21, a1)
            T = -np.trace(np.dot(v12_a2, v21_a1))
            T_e[e] = T
            self.T_e = T_e
        return T_e



    def get_Vts(self, tip_pos):
        # tip_pos is in terms of grid points, showing the tip apex position respect to the surface grid
        print('tip_pos is', tip_pos)
        tip_apex_c = self.tip_apex_c[0] 
        srf_pos_av = self.srf.atoms.get_positions() / Bohr
        srf_zmax = srf_pos_av[:,2].max()
        srf_zmax_c = np.round(srf_zmax / self.srf.gd.h_c[2]).astype(int)
        srf_ox = self.srf_extension[0] + tip_pos[0]
        srf_oy = self.srf_extension[1] + tip_pos[1]
        srf_oz = srf_zmax_c + tip_pos[2]
        shift = np.array([srf_ox, srf_oy, srf_oz])
        shift -= tip_apex_c
        for tbox in self.tip_coupling_functions:
            tbox.corner_c += shift
        for tbox in self.tip_coupling_kinetics:
            tbox.corner_c += shift
        # get the combined effective potential on the extended surface grid
        """
        This part needs to be updated for a separate xc calculation
        """
        tip_ox = self.ox
        tip_oy = self.oy
        tip_oz = self.oz
        srf_sx = srf_ox - self.nmx
        srf_ex = srf_ox + self.npx
        srf_sy = srf_oy - self.nmy
        srf_ey = srf_oy + self.npy
        srf_sz = srf_oz - self.nmz
        srf_ez = srf_oz + self.npz
        tip_sx = tip_ox - self.nmx
        tip_ex = tip_ox + self.npx
        tip_sy = tip_oy - self.nmy
        tip_ey = tip_oy + self.npy
        tip_sz = tip_oz - self.nmz
        tip_ez = tip_oz + self.npz
        V_tip = self.tip.get_effective_potential() - self.tip_alignv 
        V_couple = self.extvt_G.copy()
        V_couple[srf_sx:srf_ex, srf_sy:srf_ey, srf_sz:srf_ez] = V_tip[tip_sx:tip_ex, tip_sy:tip_ey, tip_sz:tip_ez]

        #get V_ts 
        V_ts = np.zeros((self.ni,self.nj))
        for ns in range(len(self.srf_coupling_functions)):
            sf = self.srf_coupling_functions[ns]
            sk = self.srf_coupling_kinetics[ns]
            j1 = sf[0].index
            j2 = j1 + len(sf[0])
            for t in self.tip_coupling_functions:
                i1 = t.index
                i2 = i1 + len(t)
                test_overlap1 = []
                test_overlap2 = []
                for x in range(9):
                    overlap = (t | sf[x])                    
                    if overlap is not None:
                        test_overlap1.append(overlap.copy())
                        test_overlap2.append((np.abs(overlap)).sum())
                    else:
                        test_overlap1.append(overlap)
                        test_overlap2.append(overlap)
                k = np.argmax(test_overlap2)
                if test_overlap1[k] is not None:
                    V_ts[i1:i2, j1:j2] += (t|sk[k])
                    V_ts[i1:i2, j1:j2] += (t|V_couple|sf[k])

        # return the tbox back :) 
        for tbox in self.tip_coupling_functions:
            tbox.corner_c -= shift
        for tbox in self.tip_coupling_kinetics:
            tbox.corner_c -= shift
       
        #print 'np.shape(V_ts)',np.shape(V_ts)
        return V_ts       

        
    def setup_stm(self):
        tip_pos_av = self.tip.atoms.get_positions() / Bohr
        srf_pos_av = self.srf.atoms.get_positions() / Bohr
        
        # get selected tip functions on small grids and apply kinetics
        self.tip_coupling_functions = []
        self.tip_coupling_kinetics = []
        i = 0
        for a in self.tip_coupling_atoms:
            setup = self.tip.wfs.setups[a]
            spos_c = tip_pos_av[a] / self.tip.gd.cell_c            #gives the scaled position of tip atoms in tip cell
            for phit in setup.phit_j:
                f = AtomCenteredFunctions(self.tip.gd, [phit], spos_c, i)
                f_kinetic = f.apply_t()
                self.tip_coupling_functions.append(f)
                self.tip_coupling_kinetics.append(f_kinetic)
                i += len(f.f_iG)
        self.ni = i
        #print self.tip_coupling_functions[1].size_c
        #print self.tip_coupling_functions[1].corner_c
        #print np.shape(self.tip_coupling_functions[1].f_iG)

        # get tip cutoff and surface extension
        tip_corner_c = []
        temp_min = self.tip_coupling_kinetics[0].corner_c
        temp_max = self.tip_coupling_kinetics[0].corner_c
        for f in self.tip_coupling_kinetics:
                start_c = np.minimum(f.corner_c, temp_min)
                end_c = np.maximum(f.corner_c+f.size_c, temp_max)
                temp_min = start_c
                temp_max = end_c
        #print "start_c",start_c         # in terms of grid points
        #print "end_c",end_c             # in terms of grid points
        self.tip_zmin = tip_pos_av[:,2].min()
        self.tip_apex_index = np.where(tip_pos_av[:,2] < self.tip_zmin + 0.01)
        self.tip_apex_c = np.round(tip_pos_av[self.tip_apex_index] / self.tgd.h_c).astype(int)
        self.ox, self.oy, self.oz = self.tip_apex_c[0]
        self.npx, self.npy, self.npz = np.round(end_c).astype(int) - self.tip_apex_c[0]
        self.nmx, self.nmy, self.nmz = self.tip_apex_c[0] - np.round(start_c).astype(int)

        # get the extension of srf basis functions and effective potentials
        srf_extension = np.array([self.srf.gd.N_c[0], self.srf.gd.N_c[1], 0])
        self.srf_extension = srf_extension
        print('extension of surface in terms of grids:', srf_extension)
        
        # get surface fucntions on small grid with periodic conditions with entension 
        self.srf_coupling_functions = []
        self.srf_coupling_kinetics = []
        j = 0
        for a in self.srf_coupling_atoms: 
            setup = self.srf.wfs.setups[a]
            spos_c = srf_pos_av[a] / self.srf.gd.cell_c
            for phit in setup.phit_j:
                f = AtomCenteredFunctions(self.srf.gd, [phit], spos_c, j)
                f.corner_c += srf_extension
                f_kinetic = f.apply_t()
                self.srf_coupling_functions.append(f.periodic())
                self.srf_coupling_kinetics.append(f_kinetic.periodic())
                j += len(f.f_iG)
        self.nj = j
        #print np.shape(self.srf_coupling_functions[0][0].f_iG), self.srf_coupling_functions[0][0].corner_c
        #print np.shape(self.srf_coupling_functions[0][8].f_iG), self.srf_coupling_functions[0][8].corner_c
        # Extend the surface grid and surface effective potential
        svt_G = self.srf.get_effective_potential() - self.srf_alignv + self.bias 
        ex,ey,ez = srf_extension
        newsize_c = 2 * srf_extension + self.sgd.N_c
        extvt_G = np.zeros(newsize_c)
        extvt_G[ex:-ex,ey:-ey,:] = svt_G
        extvt_G[:ex,ey:-ey,:]  = svt_G[-ex:,:,:]
        extvt_G[-ex:,ey:-ey,:] = svt_G[:ex,:,:]
        extvt_G[:,:ey,:] = extvt_G[:,-2*ey:-ey,:]
        extvt_G[:,-ey:,:] = extvt_G[:,ey:2*ey,:]
        self.extvt_G = extvt_G  

        extsgd = GridDescriptor(N_c=newsize_c,
                                cell_cv=self.sgd.h_c*(newsize_c),
                                pbc_c=False,
                                comm=mpi.serial_comm)

        ## Transfer the coner_c of surface basis-functions and srf_kinetics to the extended surface cell
        #for sbox in self.srf_coupling_functions:
        #    sbox.corner_c += srf_extension
        #for sbox in self.srf_coupling_kinetics:
        #    sbox.corner_c += srf_extension



    def align_v(self,calc1,index1,calc2,index2):
        pos1 = calc1.atoms.positions[index1]
        pos1_c = np.round(pos1/Bohr/calc1.wfs.gd.h_c).astype(int)
        pos1_v = calc1.get_effective_potential()[pos1_c[0],pos1_c[1],pos1_c[2]]
        pos2 = calc2.atoms.positions[index2]
        pos2_c = np.round(pos2/Bohr/calc2.wfs.gd.h_c).astype(int)
        pos2_v = calc2.get_effective_potential()[pos2_c[0],pos2_c[1],pos2_c[2]]
        return pos2_v - pos1_v


class Lead_HS:
    def __init__(self, calc, atom_list, extension = None, shift = 0.0):
        self.calc = calc
        self.gd = calc.wfs.gd
        self.atom_list = atom_list
        self.shift = shift
        self.extension = extension
        if self.extension is None:
            self.extension = self.gd.N_c     ####################### needs to be modified later #######################
 
    def get_HS(self):
        # get functions on small grid without periodic conditions
        self.basis_functions_fix = []
        pos_av = self.calc.atoms.get_positions() / Bohr

        # get functions on small grid with periodic conditions without entension 
        self.basis_functions = []
        self.basis_kinetics = []
        Hi = 0
        for a in self.atom_list:
            setup = self.calc.wfs.setups[a]
            spos_c = pos_av[a] / self.gd.cell_c
            for phit in setup.phit_j:
                f = AtomCenteredFunctions(self.gd, [phit], spos_c, Hi)
                f.corner_c += self.extension
                f_kinetic = f.apply_t()
                self.basis_functions_fix.append(f)
                self.basis_functions.append(f.periodic())
                self.basis_kinetics.append(f_kinetic.periodic())
                Hi += len(f.f_iG)
        self.nHi = Hi
        #print np.shape(self.srfH_kinetics[0].f_iG)
        # extend the effective potential for surface
        svt_G = petential = self.calc.get_effective_potential() - self.shift
        extension = self.extension
        ex,ey,ez = extension
        newsize_c = 2 * extension + self.gd.N_c
        ext_svt_G = np.zeros(newsize_c)
        ext_svt_G[ex:-ex,ey:-ey,ez:-ez] = svt_G
        ext_svt_G[:ex,ey:-ey,ez:-ez]  = svt_G[-ex:,:,:]
        ext_svt_G[-ex:,ey:-ey,ez:-ez] = svt_G[:ex,:,:]
        ext_svt_G[:,:ey,ez:-ez] = ext_svt_G[:,-2*ey:-ey,ez:-ez]
        ext_svt_G[:,-ey:,ez:-ez] = ext_svt_G[:,ey:2*ey,ez:-ez]
        ext_svt_G[:,:,:ez] = ext_svt_G[:,:,-2*ez:-ez]
        ext_svt_G[:,:,-ez:] = ext_svt_G[:,:,ez:2*ez]

        # get H 
        self.H = np.zeros((self.nHi, self.nHi))
        self.S = np.zeros((self.nHi, self.nHi))
        for ns in range(len(self.basis_functions)):
            sf = self.basis_functions[ns]
            sk = self.basis_kinetics[ns]
            j1 = sf[0].index
            j2 = j1 + len(sf[0])

            for s in self.basis_functions_fix:
                i1 = s.index
                i2 = i1 + len(s)
                test_overlap1 = []
                test_overlap2 = []

                for x in range(9):
                    overlap = (s | sf[x])
                    if overlap is not None:
                        test_overlap1.append(overlap.copy())
                        test_overlap2.append((np.abs(overlap)).sum())
                    else:
                        test_overlap1.append(overlap)
                        test_overlap2.append(overlap)
                k = np.argmax(test_overlap2)
                if test_overlap1[k] is not None:
                    self.H[i1:i2, j1:j2] += (s|sk[k])
                    self.H[i1:i2, j1:j2] += (s| ext_svt_G |sf[k])
                    self.S[i1:i2, j1:j2] += (s|sf[k])

        return self.H, self.S




from __future__ import print_function
from gpaw.xc.libxc import LibXC
from math import pi
import numpy as np

nspins = 1
for name in [
    'LDA', 'PBE', 'revPBE', 'RPBE',
    'LDA_X', 'GGA_X_PBE_R', 'GGA_X_RPBE',
    'LDA_C_PW',
    ]:
    xc = LibXC(name)
    xc.initialize(nspins)
    libxc = xc.xc
    lxc_fxc = libxc.calculate_fxc_spinpaired
    lxc_fxc_fd = libxc.calculate_fxc_fd_spinpaired
    na = 2.0
    if nspins == 2:
        nb = 1.0
    else:
        nb = 0.0
    print(na, nb)
    if (nb > 0.0): assert (nspins == 2)
    if nspins == 2:
        sigma0 = 2.0 # (0.0, 1.0, 1.0)
        sigma1 = 2.0
        sigma2 = 5.0 # (1.0, 2.0, 0.0)
    else:
        sigma0 = 2.0 # (0.0, 1.0, 1.0)
        sigma1 = 0.0
        sigma2 = 0.0
    taua=(3.*pi**2)**(2./3.)*na**(5./3.)/2.*sigma0
    taub=(3.*pi**2)**(2./3.)*nb**(5./3.)/2.*sigma2
    if ((sigma1 > 0.0) or (sigma2 > 0.0)): assert (nspins == 2)

    na = np.array([na])
    sigma0 = np.array([sigma0])

    dvdn = np.zeros((1))
    dvdnda2 = np.zeros((6))
    dvda2da2 = np.zeros_like(dvdnda2)

    dvdn_N = np.zeros_like(dvdn)
    dvdnda2_N = np.zeros_like(dvdnda2)
    dvda2da2_N = np.zeros_like(dvdnda2)

    lxc_fxc(na, dvdn, sigma0, dvdnda2, dvda2da2)
    lxc_fxc_fd(na, dvdn_N, sigma0, dvdnda2_N, dvda2da2_N)

    error = [0.0, 'exact']
    for E in [
        ('dvdn', dvdn[0], dvdn_N[0]),
        ('dvdnda2', dvdnda2[0], dvdnda2_N[0]),
        ('dvda2da2', dvda2da2[0], dvda2da2_N[0]),
        ]:
        for e in E[2:]:
            de = abs(e - E[1])
            if de > error[0]:
                error[0] = de
                error[1] = E[0]
    print(name, error[0], error[1])
    assert error[0] < 5.0e-7



def agts(queue):
    molecules = [queue.add('molecules.py %d' % i,
                           ncpus=1,
                           walltime=3*60)
                 for i in range(2)]
    queue.add('check.py', deps=molecules)
 



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""Occupation number objects."""

import warnings
from math import pi

import numpy as np
from ase.units import Hartree

from gpaw.utilities import erf


def findroot(func, x, tol=1e-10):
    """Function used for locating Fermi level."""
    xmin = -np.inf
    xmax = np.inf
        
    # Try 10 step using the gradient:
    niter = 0
    while True:
        f, dfdx = func(x)
        if abs(f) < tol:
            return x, niter
        if f < 0.0 and x > xmin:
            xmin = x
        elif f > 0.0 and x < xmax:
            xmax = x
        dx = -f / max(dfdx, 1e-18)
        if niter == 10 or abs(dx) > 0.01 or not (xmin < x + dx < xmax):
            break  # try bisection
        x += dx
        niter += 1

    # Bracket the solution:
    if not np.isfinite(xmin):
        xmin = x
        fmin = f
        step = 0.01
        while fmin > tol:
            xmin -= step
            fmin = func(xmin)[0]
            step *= 2
            
    if not np.isfinite(xmax):
        xmax = x
        fmax = f
        step = 0.01
        while fmax < 0:
            xmax += step
            fmax = func(xmax)[0]
            step *= 2
            
    # Bisect:
    while True:
        x = (xmin + xmax) / 2
        f = func(x)[0]
        if abs(f) < tol:
            return x, niter
        if f > 0:
            xmax = x
        else:
            xmin = x
        niter += 1


class OccupationNumbers:
    """Base class for all occupation number objects."""
    def __init__(self, fixmagmom):
        self.fixmagmom = fixmagmom
        self.magmom = None      # magnetic moment
        self.e_entropy = None   # -ST
        self.e_band = None      # band energy (sum_n eps_n * f_n)
        self.fermilevel = None  # Fermi level
        self.homo = np.nan      # HOMO eigenvalue
        self.lumo = np.nan      # LUMO eigenvalue
        self.nvalence = None    # number of electrons
        self.split = 0.0        # splitting of Fermi levels from fixmagmom=True
        self.niter = 0          # number of iterations for finding Fermi level
        self.ready = False

    def is_ready(self):
        return self.ready
        
    def calculate(self, wfs):
        """Calculate everything.

        The following is calculated:

        * occupation numbers
        * magnetic moment
        * entropy
        * band energy
        * Fermi level
        * HOMO and LUMO energies
        """

        # Allow subclasses to adjust nvalence:
        self.set_number_of_electrons(wfs)
        
        # Allocate:
        for kpt in wfs.kpt_u:
            if kpt.f_n is None:
                kpt.f_n = wfs.bd.empty()

            # There are no eigenvalues, might as well return
            if kpt.eps_n is None:
                return

            # Sanity check.  This class will typically be the first to
            # suffer if any NaNs sneak in.
            assert not np.isnan(kpt.eps_n).any()

        # Let the master domain do the work and broadcast results:
        data = np.empty(7)
        if wfs.gd.comm.rank == 0:
            self.calculate_occupation_numbers(wfs)
            self.calculate_band_energy(wfs)
            data[:] = [self.magmom, self.e_entropy, self.e_band,
                       self.homo, self.lumo,
                       self.fermilevel, self.split]
        wfs.world.broadcast(data, 0)
        (self.magmom, self.e_entropy, self.e_band,
         self.homo, self.lumo, self.fermilevel, self.split) = data

        for kpt in wfs.kpt_u:
            wfs.gd.comm.broadcast(kpt.f_n, 0)

    def set_number_of_electrons(self, wfs):
        self.nvalence = wfs.nvalence
        self.ready = True

    def calculate_occupation_numbers(self, wfs):
        raise NotImplementedError

    def calculate_band_energy(self, wfs):
        """Sum up all eigenvalues weighted with occupation numbers"""
        e_band = 0.0
        for kpt in wfs.kpt_u:
            e_band += np.dot(kpt.f_n, kpt.eps_n)
        self.e_band = wfs.kptband_comm.sum(e_band)

    def print_fermi_level(self, stream):
        pass

    def get_fermi_level(self):
        raise ValueError('Can not calculate Fermi level!')

    def set_fermi_level(self, fermilevel):
        """This method sets the fermi-level.
        
        However, since you get two fermi-levels when doing
        calculations with fixed magmom, you should be able
        to set two fermi-levels.

        So set_fermi_level will give an warning if you supply
        one fermi-level to an fixed-magmom calculation and will
        also accept everything which can converted to an numpy-
        array for setting two fermi-levels.

        You can (and should) use "set_fermi_levels" when doing
        fixed-magmom calculations which raises an error if you
        supply one fermi-level.

        For historical (and storage) reasons there is also
        an method "set_fermi_levels_mean" which might be used
        to set the fermi-levels using the mean and the splitting
        (set_fermi_splitting).

        However: you can use simply this method but have to
        keep in mind to supply two fermi-levels if you do fixed-
        magmom calculations.
            
        """
        if self.fixmagmom:
            fermilevels = np.array(fermilevel)
            if fermilevels.size == 2:
                self.fermilevel = fermilevels.mean()
                self.split = fermilevels[0] - fermilevels[1]
            else:
                warnings.warn('Please use set_fermi_levels when ' +
                              'using fixmagmom', DeprecationWarning)
                self.fermilevel = fermilevel
        else:
            self.fermilevel = fermilevel

    def set_fermi_levels(self, fermilevels):
        """This method sets two fermi levels.
        
        It takes two fermi-levels as argument and sets the
        fermi-levels in a fixed-magmom calculation according
        to the supplied levels.
        
        """
        if self.fixmagmom:
            fermilevels = np.array(fermilevels)
            if fermilevels.size == 2:
                self.fermilevel = fermilevels.mean()
                self.split = fermilevels[0] - fermilevels[1]
            else:
                raise ValueError('Please supply two distinct values.')
        else:
            raise ValueError('Different fermi levels are only vaild with ' +
                             'fixmagmom!')
            
    def set_fermi_levels_mean(self, fermilevel):
        """This method sets the mean of two fermi level.
        
        It is mostly used for storage/historical reasons.
        With the combination of this method and "set_fermi_splitting",
        you're able to set the corresponding values direct.
        
        """
        if self.fixmagmom:
            if isinstance(fermilevel, float):
                self.fermilevel = fermilevel
            else:
                raise ValueError('Please use float for supplying mean ' +
                                 'fermi level!')
        else:
            raise ValueError('Different fermi levels are only vaild with ' +
                             'fixmagmom!')
            
    def set_fermi_splitting(self, fermisplit):
        """Set the splitting of the fermi-level (in Ht).
        
        This method is used in fixed-magmom calculations.
        
        """
        self.split = fermisplit


class ZeroKelvin(OccupationNumbers):
    def __init__(self, fixmagmom):
        self.width = 0.0
        OccupationNumbers.__init__(self, fixmagmom)
        
    def calculate_occupation_numbers(self, wfs):
        if wfs.nspins == 1:
            self.spin_paired(wfs)
        elif self.fixmagmom:
            assert wfs.kd.gamma
            self.fixed_moment(wfs)
        else:
            assert wfs.kd.nibzkpts == 1
            self.spin_polarized(wfs)

        self.e_entropy = 0.0

    def occupy(self, f_n, eps_n, ne, weight=1):
        """Fill in occupation numbers.

        return HOMO and LUMO energies."""
        N = len(f_n)
        if ne == N * weight:
            f_n[:] = weight
            return eps_n[-1], np.inf

        n, f = divmod(ne, weight)
        n = int(n)
        f_n[:n] = weight
        assert n < N
        f_n[n] = f
        f_n[n + 1:] = 0.0
        if f > 0.0:
            return eps_n[n], eps_n[n]
        return eps_n[n - 1], eps_n[n]

    def print_fermi_level(self, stream):
        if self.fermilevel is not None and np.isfinite(self.fermilevel):
            # if self.split == 0.0:
            if not self.fixmagmom:
                stream.write('Fermi Level: %.5f\n' %
                             (Hartree * self.fermilevel))
            else:
                stream.write('Fermi Levels: %.5f, %.5f\n' %
                             (Hartree * (self.fermilevel + 0.5 * self.split),
                              Hartree * (self.fermilevel - 0.5 * self.split)))

    def get_fermi_level(self):
        """This function returns the calculated fermi-level.

        Care: you get two distinct fermi-levels if you do
        fixed-magmom calculations. Therefore you should use
        "get_fermi_levels" or "get_fermi_levels_mean" in
        conjunction with "get_fermi_splitting" if you do
        fixed-magmom calculations. We will issue an warning
        otherwise.
        
        """
        if self.fermilevel is None or not np.isfinite(self.fermilevel):
            OccupationNumbers.get_fermi_level(self)  # fail
        else:
            if self.fixmagmom:
                warnings.warn('Please use get_fermi_levels when ' +
                              'using fixmagmom', DeprecationWarning)
                fermilevels = np.empty(2)
                fermilevels[0] = self.fermilevel + 0.5 * self.split
                fermilevels[1] = self.fermilevel - 0.5 * self.split
                return fermilevels
            else:
                return self.fermilevel

    def get_fermi_levels(self):
        """Getting fermi-levels in case of fixed-magmom."""
        if self.fermilevel is None or not np.isfinite(self.fermilevel):
            OccupationNumbers.get_fermi_level(self)  # fail
        else:
            if self.fixmagmom:
                fermilevels = np.empty(2)
                fermilevels[0] = self.fermilevel + 0.5 * self.split
                fermilevels[1] = self.fermilevel - 0.5 * self.split
                return fermilevels
            else:
                raise ValueError('Distinct fermi-levels are only vaild ' +
                                 'for fixed-magmom calculations!')

    def get_fermi_levels_mean(self):
        if self.fermilevel is None or not np.isfinite(self.fermilevel):
            OccupationNumbers.get_fermi_level(self)  # fail
        else:
            return self.fermilevel

    def get_fermi_splitting(self):
        """Return the splitting of the fermi level in hartree.
            
        Returns 0.0 if calculation is not done using
        fixmagmom.

        """
        if self.fixmagmom:
            return self.split
        else:
            return 0.0

    def get_homo_lumo(self, wfs):
        if self.nvalence is None:
            self.calculate(wfs)
        if np.isfinite(self.homo) and np.isfinite(self.lumo):
            if wfs.bd.comm.rank != 0:
                homolumo = np.array([0.0, 0.0])
            else:
                homolumo = np.array([self.homo, self.lumo])
            wfs.bd.comm.broadcast(homolumo, 0)
            return homolumo
        else:
            raise ValueError("Can't find HOMO and/or LUMO!")

    def fixed_moment(self, wfs):
        assert wfs.nspins == 2 and wfs.kd.nbzkpts == 1
        fermilevels = np.zeros(2)
        for kpt in wfs.kpt_u:
            eps_n = wfs.bd.collect(kpt.eps_n)
            if eps_n is None:
                f_n = None
            else:
                f_n = wfs.bd.empty(global_array=True)
                sign = 1 - kpt.s * 2
                ne = 0.5 * (self.nvalence + sign * self.magmom)

                homo, lumo = self.occupy(f_n, eps_n, ne)

                fermilevels[kpt.s] = 0.5 * (homo + lumo)
            wfs.bd.distribute(f_n, kpt.f_n)
        wfs.kptband_comm.sum(fermilevels)
        self.fermilevel = fermilevels.mean()
        self.split = fermilevels[0] - fermilevels[1]
        
    def spin_paired(self, wfs):
        self.homo = -np.inf
        self.lumo = np.inf
        for kpt in wfs.kpt_u:
            eps_n = wfs.bd.collect(kpt.eps_n)
            if wfs.bd.comm.rank == 0:
                f_n = wfs.bd.empty(global_array=True)
                homo, lumo = self.occupy(f_n, eps_n,
                                         0.5 * self.nvalence * wfs.ncomp *
                                         kpt.weight, kpt.weight)
                self.homo = max(self.homo, homo)
                self.lumo = min(self.lumo, lumo)
            else:
                f_n = None
                self.fermilevel = np.nan
            wfs.bd.distribute(f_n, kpt.f_n)

        if wfs.bd.comm.rank == 0:
            self.homo = wfs.kd.comm.max(self.homo)
            self.lumo = wfs.kd.comm.min(self.lumo)
            self.fermilevel = 0.5 * (self.homo + self.lumo)

        self.magmom = 0.0
        
    def spin_polarized(self, wfs):
        eps_un = [wfs.bd.collect(kpt.eps_n) for kpt in wfs.kpt_u]
        self.fermilevel = np.nan
        nbands = wfs.bd.nbands
        if wfs.bd.comm.rank == 0:
            if wfs.kd.comm.size == 2:
                if wfs.kd.comm.rank == 1:
                    wfs.kd.comm.send(eps_un[0], 0)
                else:
                    eps_sn = [eps_un[0], np.empty(nbands)]
                    wfs.kd.comm.receive(eps_sn[1], 1)
            else:
                eps_sn = eps_un

            if wfs.kd.comm.rank == 0:
                eps_n = np.ravel(eps_sn)
                f_n = np.empty(nbands * 2)
                nsorted = eps_n.argsort()
                self.homo, self.lumo = self.occupy(f_n, eps_n[nsorted],
                                                   self.nvalence)
                f_sn = f_n[nsorted.argsort()].reshape((2, nbands))
                self.magmom = f_sn[0].sum() - f_sn[1].sum()
                self.fermilevel = 0.5 * (self.homo + self.lumo)

            if wfs.kd.comm.size == 2:
                if wfs.kd.comm.rank == 0:
                    wfs.kd.comm.send(f_sn[1], 1)
                else:
                    f_sn = [None, np.empty(nbands)]
                    wfs.kd.comm.receive(f_sn[1], 0)
        else:
            f_sn = [None, None]

        for kpt in wfs.kpt_u:
            wfs.bd.distribute(f_sn[kpt.s], kpt.f_n)


class SmoothDistribution(ZeroKelvin):
    """Base class for Fermi-Dirac and other smooth distributions."""
    def __init__(self, width, fixmagmom, maxiter):
        """Smooth distribution.

        Find the Fermi level by integrating in energy until
        the number of electrons is correct.

        width: float
            Width of distribution in eV.
        fixmagmom: bool
            Fix spin moment calculations.  A separate Fermi level for
            spin up and down electrons is found: self.fermilevel +
            self.split and self.fermilevel - self.split.
        maxiter: int
            Maximum number of iterations.
        """

        ZeroKelvin.__init__(self, fixmagmom)
        self.width = width / Hartree
        self.maxiter = maxiter
        
    def __repr__(self):
        classname = self.__class__.__name__
        template = '%s(width=%f, fixmagmom=%r, maxiter=%d)'
        string = template % (classname, self.width * Hartree, self.fixmagmom,
                             self.maxiter)
        return string

    def calculate_occupation_numbers(self, wfs):
        if self.width == 0 or self.nvalence == wfs.bd.nbands * 2 // wfs.ncomp:
            ZeroKelvin.calculate_occupation_numbers(self, wfs)
            return

        if self.fermilevel is None:
            self.fermilevel = self.guess_fermi_level(wfs)

        if not self.fixmagmom or wfs.nspins == 1:
            result = self.find_fermi_level(wfs, self.nvalence, self.fermilevel)
            self.fermilevel, self.magmom, self.e_entropy = result
            
            if wfs.nspins == 1:
                self.magmom = 0.0
        else:
            fermilevels = np.empty(2)
            self.e_entropy = 0.0
            for s in range(2):
                sign = 1 - s * 2
                ne = 0.5 * (self.nvalence + sign * self.magmom)
                fermilevel = self.fermilevel + 0.5 * sign * self.split
                fermilevels[s], magmom, e_entropy = \
                    self.find_fermi_level(wfs, ne, fermilevel, [s])
                self.e_entropy += e_entropy
            self.fermilevel = fermilevels.mean()
            self.split = fermilevels[0] - fermilevels[1]

    def get_homo_lumo_by_spin(self, wfs, spin):
        if wfs.nspins == 1:
            assert spin == 0
            n = self.nvalence // 2
            band_rank, myn = wfs.bd.who_has(n - 1)
            if wfs.bd.rank == band_rank:
                homo = max([kpt.eps_n[myn] for kpt in wfs.kpt_u])
            else:
                homo = -1000.0
            homo = wfs.world.max(homo)

            # There are not enough bands for LUMO
            if n >= wfs.bd.nbands:
                return np.array([homo, np.NaN])

            band_rank, myn = wfs.bd.who_has(n)
            if wfs.bd.rank == band_rank:
                lumo = -max([-kpt.eps_n[myn] for kpt in wfs.kpt_u])
            else:
                lumo = 1000.0
            lumo = -wfs.world.max(-lumo)
            return np.array([homo, lumo])
        else:
            assert wfs.bd.comm.size == 1
            eps_homo = -1000.0
            eps_lumo = 1000.0
            epsilon = 1e-2
            for kpt in wfs.kpt_u:
                if kpt.s == spin:
                    eps = np.max(np.where(kpt.f_n / kpt.weight > epsilon,
                                          kpt.eps_n, -1000.0))
                    eps_homo = max([eps_homo, eps])
                    eps = np.min(np.where(kpt.f_n / kpt.weight <= epsilon,
                                          kpt.eps_n, +1000.0))
                    eps_lumo = min([eps_lumo, eps])
            
            eps_homo = wfs.kd.comm.max(eps_homo)
            eps_lumo = wfs.kd.comm.min(eps_lumo)

            return np.array([eps_homo, eps_lumo])

    def get_homo_lumo(self, wfs):
        if self.width == 0:
            return ZeroKelvin.get_homo_lumo(self, wfs)

        if wfs.nspins == 2:
            raise NotImplementedError

        if self.nvalence is None:
            self.calculate(wfs)
        
        return self.get_homo_lumo_by_spin(wfs, 0)

    def guess_fermi_level(self, wfs):
        fermilevel = 0.0

        kd = wfs.kd
        
        myeps_un = np.empty((kd.mynks, wfs.bd.nbands))
        for u, kpt in enumerate(wfs.kpt_u):
            myeps_un[u] = wfs.bd.collect(kpt.eps_n)
        
        if wfs.bd.comm.rank == 0:
            eps_skn = kd.collect(myeps_un, broadcast=False)
            if kd.comm.rank == 0:
                eps_n = eps_skn.ravel()
                w_skn = np.empty((kd.nspins, kd.nibzkpts, wfs.bd.nbands))
                w_skn[:] = (2.0 / wfs.nspins / wfs.ncomp *
                            kd.weight_k[:, np.newaxis])
                w_n = w_skn.ravel()
                n_i = eps_n.argsort()
                w_i = w_n[n_i]
                f_i = np.add.accumulate(w_i) - 0.5 * w_i
                i = np.nonzero(f_i >= self.nvalence)[0][0]
                if i == 0:
                    fermilevel = eps_n[n_i[0]]
                else:
                    fermilevel = ((eps_n[n_i[i]] *
                                   (self.nvalence - f_i[i - 1]) +
                                   eps_n[n_i[i - 1]] *
                                   (f_i[i] - self.nvalence)) /
                                  (f_i[i] - f_i[i - 1]))

        # XXX broadcast would be better!
        return wfs.kptband_comm.sum(fermilevel)
                    
    def find_fermi_level(self, wfs, ne, fermilevel, spins=(0, 1)):
        niter = 0
        
        x = self.fermilevel
        if x is None:
            x = self.guess_fermi_level(wfs)
            
        data = np.empty(4)
        
        def f(x, data=data):
            data.fill(0.0)
            for kpt in wfs.kpt_u:
                if kpt.s in spins:
                    data += self.distribution(kpt, x)
            wfs.kptband_comm.sum(data)
            n, dnde = data[:2]
            dn = n - ne
            return dn, dnde

        fermilevel, niter = findroot(f, x)
        
        self.niter = niter
        magmom, e_entropy = data[2:]
        return fermilevel, magmom, e_entropy


class FermiDirac(SmoothDistribution):
    def __init__(self, width, fixmagmom=False, maxiter=10000):
        SmoothDistribution.__init__(self, width, fixmagmom, maxiter)

    def distribution(self, kpt, fermilevel):
        x = (kpt.eps_n - fermilevel) / self.width
        x = x.clip(-100, 100)
        y = np.exp(x)
        z = y + 1.0
        kpt.f_n[:] = kpt.weight / z
        n = kpt.f_n.sum()
        dnde = (n - (kpt.f_n**2).sum() / kpt.weight) / self.width
        y *= x
        y /= z
        y -= np.log(z)
        e_entropy = -kpt.weight * y.sum() * self.width
        sign = 1 - kpt.s * 2
        return np.array([n, dnde, n * sign, e_entropy])

        
class MethfesselPaxton(SmoothDistribution):
    def __init__(self, width, iter=0, fixmagmom=False, maxiter=1000):
        SmoothDistribution.__init__(self, width, fixmagmom, maxiter)
        self.iter = iter

    def distribution(self, kpt, fermilevel):
        x = (kpt.eps_n - fermilevel) / self.width
        x = x.clip(-100, 100)

        z = 0.5 * (1 - erf(x))
        for i in range(self.iter):
            z += (self.coff_function(i + 1) *
                  self.hermite_poly(2 * i + 1, x) * np.exp(-x**2))
        kpt.f_n[:] = kpt.weight * z
        n = kpt.f_n.sum()

        dnde = 1 / np.sqrt(pi) * np.exp(-x**2)
        for i in range(self.iter):
            dnde += (self.coff_function(i + 1) *
                     self.hermite_poly(2 * i + 2, x) * np.exp(-x**2))
        dnde = dnde.sum()
        dnde *= kpt.weight / self.width
        e_entropy = (0.5 * self.coff_function(self.iter) *
                     self.hermite_poly(2 * self.iter, x) * np.exp(-x**2))
        e_entropy = kpt.weight * e_entropy.sum() * self.width

        sign = 1 - kpt.s * 2
        return np.array([n, dnde, n * sign, e_entropy])

    def coff_function(self, n):
        return (-1)**n / (np.product(np.arange(1, n + 1)) *
                          4**n * np.sqrt(np.pi))

    def hermite_poly(self, n, x):
        if n == 0:
            return 1
        elif n == 1:
            return 2 * x
        else:
            return (2 * x * self.hermite_poly(n - 1, x) -
                    2 * (n - 1) * self.hermite_poly(n - 2, x))

                            
class FixedOccupations(ZeroKelvin):
    def __init__(self, occupation):
        self.occupation = np.array(occupation)
        ZeroKelvin.__init__(self, True)

    def spin_paired(self, wfs):
        return self.fixed_moment(wfs)

    def fixed_moment(self, wfs):
        for kpt in wfs.kpt_u:
            wfs.bd.distribute(self.occupation[kpt.s], kpt.f_n)


class TFOccupations(FermiDirac):
    def __init__(self, width, fixmagmom=False, maxiter=1000):
        FermiDirac.__init__(self, width, fixmagmom, maxiter)
    
    def occupy(self, f_n, eps_n, ne, weight=1):
        """Fill in occupation numbers.
        
        In TF mode only one band. Is guaranteed to work only
        for spin-paired case.
        
        return HOMO and LUMO energies."""
        # Same as occupy in FermiDirac expect one band: weight = ne
        return FermiDirac.occupy(self, f_n, eps_n, ne, ne)



def agts(queue):
    run = [queue.add('pbe_jacapo.py %s' % r,
                     queueopts='-l nodes=1:ppn=4:opteron4',
                     ncpus=1,
                     walltime=10*60)
           for r in range(0)]
    if 0:  # do not perform analysis
        # we keep once generated files static
        analyse = queue.add('analyse.py dcdft_pbe_jacapo',
                            ncpus=1, walltime=10, deps=run,
                            creates=['dcdft_pbe_jacapo.csv',
                                     'dcdft_pbe_jacapo.txt',
                                     'dcdft_pbe_jacapo_Delta.txt',
                                     'dcdft_pbe_jacapo_raw.csv'])



import optparse

from ase.units import Bohr, Hartree
from ase.tasks.calcfactory import CalculatorFactory, str2dict

from gpaw.utilities.gpts import get_number_of_grid_points
from gpaw.wavefunctions.pw import PW


class GPAWFactory(CalculatorFactory):
    def __init__(self, show_text_output=False, write_gpw_file=None,
                 **kwargs):
        self.show_text_output = show_text_output
        # this is used as write(mode=write_gpw_file)
        # so it should be called rather write_gpw_mode?
        if write_gpw_file is not None:
            assert isinstance(write_gpw_file, str)
        self.write_gpw_file = write_gpw_file

        CalculatorFactory.__init__(self, None, 'GPAW', **kwargs)

    def __call__(self, name, atoms):
        kpts = self.calculate_kpts(atoms)

        kwargs = self.kwargs.copy()  # modify a copy
        
        if (not atoms.pbc.any() and len(atoms) == 1 and
            atoms.get_initial_magnetic_moments().any() and
            'hund' not in kwargs):
            kwargs['hund'] = True

        if atoms.pbc.any() and 'gpts' not in kwargs:
            # Use fixed number of gpts:
            h = kwargs.get('h')
            if h is not None:
                h /= Bohr

            cell_cv = atoms.cell / Bohr

            mode = kwargs.get('mode')
            if mode == 'pw':
                mode = PW()

            gpts = get_number_of_grid_points(cell_cv, h, mode,
                                             kwargs.get('realspace'))
            kwargs['h'] = None
            kwargs['gpts'] = gpts
            
            if isinstance(mode, PW):
                kwargs['mode'] = PW(mode.ecut * Hartree,
                                    mode.fftwflags,
                                    atoms.cell)

        if self.show_text_output:
            txt = '-'
        else:
            txt = name + '.txt'


        from gpaw import GPAW
        return GPAW(txt=txt, kpts=kpts, **kwargs)
        
    def add_options(self, parser):
        CalculatorFactory.add_options(self, parser)
        
        calc = optparse.OptionGroup(parser, 'GPAW')
        calc.add_option('--parameter-file', metavar='FILE',
                        help='Read GPAW parameters from file.')
        calc.add_option('-S', '--show-text-output', action='store_true',
                        help='Send text output from calculation to ' +
                        'standard out.')
        calc.add_option('-W', '--write-gpw-file', metavar='MODE',
                        help='Write gpw file.')
        parser.add_option_group(calc)

    def parse(self, opts, args):
        if opts.parameters:
            # Import stuff that eval() may need to know:
            from gpaw.wavefunctions.pw import PW
            from gpaw.occupations import FermiDirac, MethfesselPaxton
            from gpaw.mixer import Mixer, MixerSum
            from gpaw.poisson import PoissonSolver
            from gpaw.eigensolvers import RMM_DIIS
       
            self.kwargs.update(str2dict(opts.parameters, locals()))
            opts.parameters = None

        CalculatorFactory.parse(self, opts, args)

        self.show_text_output = opts.show_text_output
        self.write_gpw_file = opts.write_gpw_file

        if opts.parameter_file:
            self.kwargs.update(eval(open(opts.parameter_file).read()))



from __future__ import print_function, division
import multiprocessing as mp
import glob
import os
import random
import re
import time
import traceback

import numpy as np
from ase import Atoms
from ase.data import covalent_radii, atomic_numbers
from ase.lattice import bulk
from ase.lattice.surface import fcc111
from ase.units import Bohr

from gpaw import GPAW, PW, setup_paths, Mixer  # , ConvergenceError
from gpaw.atom.generator2 import _generate, DatasetGenerationError


my_covalent_radii = covalent_radii.copy()
my_covalent_radii[1] += 0.2
my_covalent_radii[2] += 0.2
my_covalent_radii[4] -= 0.1  # Be
my_covalent_radii[5] -= 0.1
my_covalent_radii[6] -= 0.1
my_covalent_radii[7] -= 0.1
my_covalent_radii[8] -= 0.1
my_covalent_radii[9] -= 0.1  # F
my_covalent_radii[10] += 0.2
for e in ['Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr']:  # missing radii
    my_covalent_radii[atomic_numbers[e]] = 1.7


NN = 11


class GA:
    def __init__(self, initialvalue=None):
        self.initialvalue = initialvalue
        
        self.individuals = {}
        self.errors = {}
        
        if os.path.isfile('pool.csv'):
            for line in open('pool.csv'):
                words = line.split(',')
                n = int(words.pop(0))
                error = float(words.pop(0))
                x = tuple(int(word) for word in words[:-NN])
                self.individuals[x] = (error, n)
                y = tuple(float(word) for word in words[-NN:])
                self.errors[n] = y
                
        self.fd = open('pool.csv', 'a')  # pool of genes
        self.n = len(self.individuals)
        self.pool = mp.Pool()  # process pool

    def run(self, func, sleep=5, mutate=5.0, size1=2, size2=1000):
        results = []
        while True:
            while len(results) < mp.cpu_count():
                x = self.new(mutate, size1, size2)
                self.individuals[x] = (None, self.n)
                result = self.pool.apply_async(func, [self.n, x])
                self.n += 1
                results.append(result)
            time.sleep(sleep)
            for result in results:
                if result.ready():
                    break
            else:
                continue
            results.remove(result)
            n, x, errors, error = result.get()
            self.individuals[x] = (error, n)
            print('{0},{1},{2},{3}'.format(n, error,
                                           ','.join(str(i) for i in x),
                                           ','.join('{0:.4f}'.format(e)
                                                    for e in errors)),
                  file=self.fd)
            self.fd.flush()
            
            best = sorted(self.individuals.values())[:20]
            nbest = [N for e, N in best]
            for f in glob.glob('[0-9]*.txt'):
                if int(f[:-4]) not in nbest:
                    os.remove(f)
                    
            if len(self.individuals) > 40 and best[0][0] == np.inf:
                for result in results:
                    result.wait()
                return
                
    def new(self, mutate, size1, size2):
        if len(self.individuals) == 0:
            return self.initialvalue

        if len(self.individuals) < 32:
            x3 = np.array(self.initialvalue, dtype=float)
            mutate *= 2.5
        else:
            all = sorted((y, x)
                         for x, y in self.individuals.items()
                         if y is not None)  # and y != np.inf)
            S = len(all)
            if S < size1:
                x3 = np.array(self.initialvalue, dtype=float)
            else:
                parents = random.sample(all[:size1], 2)
                if S > size1 and random.random() < 0.33:
                    i = random.randint(size1, min(S, size2) - 1)
                    parents.append(all[i])
                    del parents[random.randint(0, 1)]
                else:
                    mutate /= 3
                    
                x1 = parents[0][1]
                x2 = parents[1][1]
                r = np.random.rand(len(x1))
                x3 = r * x1 + (1 - r) * x2

        while True:
            x3 += np.random.normal(0, mutate, len(x3))
            x = tuple(int(round(a)) for a in x3)
            if x not in self.individuals:
                break
        
        return x


def read_reference(name, symbol):
    for line in open('../../{0}.csv'.format(name)):
        words = line.split(',')
        if words[0] == 'c':
            x = [float(word) for word in words[2:]]
        elif words[0] == symbol:
            ref = dict((c, float(word))
                       for c, word in zip(x, words[2:]))
            ref['a'] = float(words[1])
            return ref
                
        
def fit(E):
    em, e0, ep = E
    a = (ep + em) / 2 - e0
    b = (ep - em) / 2
    return -b / (2 * a)
    
    
class DatasetOptimizer:
    tolerances = np.array([0.3,
                           0.01, 0.03, 0.05,
                           0.01, 0.03, 0.05,
                           40,
                           400,  # 0.1 eV convergence
                           0.002,  # eggbox error
                           0.4])
    
    conf = None
    
    def __init__(self, symbol='H'):
    
        self.symbol = symbol

        with open('../start.txt') as fd:
            for line in fd:
                words = line.split()
                if words[1] == symbol:
                    projectors = words[3]
                    radii = [float(f) for f in words[5].split(',')]
                    r0 = float(words[7].split(',')[1])
                    break
            
        # Parse projectors string:
        pattern = r'(-?\d+\.\d)'
        energies = []
        for m in re.finditer(pattern, projectors):
            energies.append(float(projectors[m.start():m.end()]))
        self.projectors = re.sub(pattern, '%.1f', projectors)
        self.nenergies = len(energies)
        
        if 'f' in self.projectors:
            self.lmax = 3
        else:
            self.lmax = 2
            
        # Round to integers:
        x = ([e / 0.1 for e in energies] +
             [r / 0.05 for r in radii] +
             [r0 / 0.05])
        self.x = tuple(int(round(f)) for f in x)
        
        # Read FHI-Aims data:
        self.reference = {'fcc': read_reference('fcc', symbol),
                          'rocksalt': read_reference('rocksalt', symbol)}

        self.ecut1 = 450.0
        self.ecut2 = 800.0
        
        setup_paths[:0] = ['../..', '.']
        
        self.Z = atomic_numbers[symbol]
        self.rc = my_covalent_radii[self.Z]
        self.rco = my_covalent_radii[8]

    def run(self):  # , mu, n1, n2):
        # mu = float(mu)
        # n1 = int(n1)
        # n2 = int(n2)
        ga = GA(self.x)
        ga.run(self)  # , mutate=mu, size1=n1, size2=n2)
        
    def best(self, N=None):
        ga = GA(self.x)
        best = sorted((error, id, x)
                      for x, (error, id) in ga.individuals.items())
        if 0:
            import pickle
            pickle.dump(sorted(ga.individuals.values()), open('Zn.pckl', 'w'))
        if N is None:
            return len(best), best[0] + (ga.errors[best[0][1]],)
        else:
            return [(error, id, x, ga.errors[id])
                    for error, id, x in best[:N]]
        
    def summary(self, N=10):
        # print('dFffRrrICEer:')
        for error, id, x, errors in self.best(N):
            params = [0.1 * p for p in x[:self.nenergies]]
            params += [0.05 * p for p in x[self.nenergies:]]
            print('{0:2} {1:2} {2:5} {3:7.1f} {4} {5}'.format(
                self.Z,
                self.symbol,
                id, error,
                ' '.join('{0:5.2f}'.format(p) for p in params),
                ' '.join('{0:8.3f}'.format(e) for e in errors)))
            
    def best1(self):
        try:
            n, (error, id, x, errors) = self.best()
        except IndexError:
            return
        energies, radii, r0, projectors = self.parameters(x)
        if 0:
            print(error, self.symbol, n)
        if 1:
            if projectors[-1].isupper():
                nderiv0 = 5
            else:
                nderiv0 = 2
            fmt = '{0:2} {1:2} -P {2:31} -r {3:20} -0 {4},{5:.2f} # {6:10.3f}'
            print(fmt.format(self.Z,
                             self.symbol,
                             projectors,
                             ','.join('{0:.2f}'.format(r) for r in radii),
                             nderiv0,
                             r0,
                             error))
        if 0:
            with open('parameters.txt', 'w') as fd:
                print(projectors, ' '.join('{0:.2f}'.format(r)
                                           for r in radii + [r0]),
                      file=fd)
        if 0 and error != np.inf and error != np.nan:
            self.generate(None, 'PBE', projectors, radii, r0, True, '',
                          logderivs=not False)
        
    def generate(self, fd, xc, projectors, radii, r0,
                 scalar_relativistic=False, tag=None, logderivs=True):
        if projectors[-1].isupper():
            nderiv0 = 5
        else:
            nderiv0 = 2
        gen = _generate(self.symbol, xc, self.conf, projectors, radii,
                        scalar_relativistic, None, r0, nderiv0,
                        ('poly', 4), None, None, fd)
        if not gen.check_all():
            raise DatasetGenerationError(xc)

        if tag is not None:
            gen.make_paw_setup(tag or None).write_xml()

        r = 1.1 * gen.rcmax
        energies = np.linspace(-1.5, 2.0, 100)
        de = energies[1] - energies[0]
        error = 0.0
        if logderivs:
            for l in range(4):
                ld1 = gen.aea.logarithmic_derivative(l, energies, r)
                ld2 = gen.logarithmic_derivative(l, energies, r)
                error = max(error, abs(ld1 - ld2).sum() * de)
        return error
            
    def parameters(self, x):
        energies = tuple(0.1 * i for i in x[:self.nenergies])
        radii = [0.05 * i for i in x[self.nenergies:-1]]
        r0 = 0.05 * x[-1]
        projectors = self.projectors % energies
        return energies, radii, r0, projectors
        
    def __call__(self, n, x):
        fd = open('{0}.txt'.format(n), 'w')
        energies, radii, r0, projectors = self.parameters(x)
        
        if any(r < r0 for r in radii):  # or any(e <= 0.0 for e in energies):
            return n, x, [np.inf] * NN, np.inf
            
        try:
            errors = self.test(n, fd, projectors, radii, r0)
        except Exception:
            traceback.print_exc(file=fd)
            errors = [np.inf] * NN
            
        try:
            os.remove('{0}.ga{1}.PBE'.format(self.symbol, n))
        except OSError:
            pass
        
        return n, x, errors, ((errors / self.tolerances)**2).sum()
        
    def test(self, n, fd, projectors, radii, r0):
        error = self.generate(fd, 'PBE', projectors, radii, r0,
                              tag='ga{0}'.format(n))
        for xc in ['PBE', 'LDA', 'PBEsol', 'RPBE', 'PW91']:
            error += self.generate(fd, xc, projectors, radii, r0,
                                   scalar_relativistic=True)
        results = {'dataset': error}
        
        for name in ['slab', 'fcc', 'rocksalt', 'convergence', 'eggbox']:
            result = getattr(self, name)(n, fd)
            results[name] = result
            
        rc = self.rc / Bohr
        results['radii'] = sum(r - rc for r in radii if r > rc)
        
        errors = self.calculate_total_error(fd, results)
        
        return errors

    def calculate_total_error(self, fd, results):
        errors = [results['dataset']]
        maxiter = results['slab']
        
        for name in ['fcc', 'rocksalt']:
            result = results[name]
            maxiter = max(maxiter, result['maxiter'])
            errors.append(result['a'] - result['a0'])
            errors.append(result['c90'] - result['c90ref'])
            errors.append(result['c80'] - result['c80ref'])
        
        errors.append(maxiter)
        errors.append(results['convergence'])
        
        errors.append(results['eggbox'])
        
        errors.append(results['radii'])
        
        return errors
        
    def fcc(self, n, fd):
        ref = self.reference['fcc']
        a0r = ref['a']  # scalar-relativistic minimum
        sc = min(0.8, 2 * self.rc * 2**0.5 / a0r)
        sc = min((abs(s - sc), s) for s in ref if s != 'a')[1]
        maxiter = 0
        energies = []
        M = 200
        mixer = {}
        if 58 <= self.Z <= 70 or 90 <= self.Z <= 102:
            M = 999
            mixer = {'mixer': Mixer(0.01, 5)}
        for s in [sc, 0.9, 0.95, 1.0, 1.05]:
            atoms = bulk(self.symbol, 'fcc', a0r * s)
            atoms.calc = GPAW(mode=PW(self.ecut2),
                              kpts={'density': 4.0, 'even': True},
                              xc='PBE',
                              lmax=self.lmax,
                              setups='ga' + str(n),
                              maxiter=M,
                              txt=fd,
                              **mixer)
            e = atoms.get_potential_energy()
            maxiter = max(maxiter, atoms.calc.get_number_of_iterations())
            energies.append(e)
                    
        return {'c90': energies[1] - energies[3],
                'c80': energies[0] - energies[3],
                'c90ref': ref[0.9] - ref[1.0],
                'c80ref': ref[sc] - ref[1.0],
                'a0': fit([ref[s] for s in [0.95, 1.0, 1.05]]) * 0.05 * a0r,
                'a': fit(energies[2:]) * 0.05 * a0r,
                'maxiter': maxiter}
        
    def rocksalt(self, n, fd):
        ref = self.reference['rocksalt']
        a0r = ref['a']
        sc = min(0.8, 2 * (self.rc + self.rco) / a0r)
        sc = min((abs(s - sc), s) for s in ref if s != 'a')[1]
        maxiter = 0
        energies = []
        M = 200
        mixer = {}
        if 58 <= self.Z <= 70 or 90 <= self.Z <= 102:
            M = 999
            mixer = {'mixer': Mixer(0.01, 5)}
        for s in [sc, 0.9, 0.95, 1.0, 1.05]:
            atoms = bulk(self.symbol + 'O', 'rocksalt', a0r * s)
            atoms.calc = GPAW(mode=PW(self.ecut2),
                              kpts={'density': 4.0, 'even': True},
                              xc='PBE',
                              lmax=self.lmax,
                              setups={self.symbol: 'ga' + str(n)},
                              maxiter=M,
                              txt=fd,
                              **mixer)
            e = atoms.get_potential_energy()
            maxiter = max(maxiter, atoms.calc.get_number_of_iterations())
            energies.append(e)
        
        return {'c90': energies[1] - energies[3],
                'c80': energies[0] - energies[3],
                'c90ref': ref[0.9] - ref[1.0],
                'c80ref': ref[sc] - ref[1.0],
                'a0': fit([ref[s] for s in [0.95, 1.0, 1.05]]) * 0.05 * a0r,
                'a': fit(energies[2:]) * 0.05 * a0r,
                'maxiter': maxiter}
        
    def slab(self, n, fd):
        a0 = self.reference['fcc']['a']
        atoms = fcc111(self.symbol, (1, 1, 7), a0, vacuum=3.5)
        assert not atoms.pbc[2]
        M = 333
        if 58 <= self.Z <= 70 or 90 <= self.Z <= 102:
            M = 1333
            mixer = {'mixer': Mixer(0.001, 3, 100)}
        else:
            mixer = {}
        atoms.calc = GPAW(mode=PW(self.ecut1),
                          kpts={'density': 2.0, 'even': True},
                          xc='PBE',
                          lmax=self.lmax,
                          setups='ga' + str(n),
                          maxiter=M,
                          txt=fd,
                          **mixer)
        atoms.get_potential_energy()
        itrs = atoms.calc.get_number_of_iterations()
        return itrs
        
    def eggbox(self, n, fd):
        h = 0.18
        a0 = 16 * h
        atoms = Atoms(self.symbol, cell=(a0, a0, a0), pbc=True)
        M = 333
        if 58 <= self.Z <= 70 or 90 <= self.Z <= 102:
            M = 999
            mixer = {'mixer': Mixer(0.01, 5)}
        else:
            mixer = {}
        atoms.calc = GPAW(h=h,
                          xc='PBE',
                          lmax=self.lmax,
                          symmetry='off',
                          setups='ga' + str(n),
                          maxiter=M,
                          txt=fd,
                          **mixer)
        e0 = atoms.get_potential_energy()
        atoms.positions += h / 6
        e1 = atoms.get_potential_energy()
        atoms.positions += h / 6
        e2 = atoms.get_potential_energy()
        atoms.positions += h / 6
        e3 = atoms.get_potential_energy()
        return np.ptp([e0, e1, e2, e3])

    def convergence(self, n, fd):
        a = 3.0
        atoms = Atoms(self.symbol, cell=(a, a, a), pbc=True)
        M = 333
        if 58 <= self.Z <= 70 or 90 <= self.Z <= 102:
            M = 999
            mixer = {'mixer': Mixer(0.01, 5)}
        else:
            mixer = {}
        atoms.calc = GPAW(mode=PW(1500),
                          xc='PBE',
                          lmax=self.lmax,
                          setups='ga' + str(n),
                          symmetry='off',
                          maxiter=M,
                          txt=fd,
                          **mixer)
        e0 = atoms.get_potential_energy()
        de0 = 0.0
        for ec in range(800, 200, -100):
            atoms.calc.set(mode=PW(ec))
            atoms.calc.set(eigensolver='rmm-diis')
            de = abs(atoms.get_potential_energy() - e0)
            if de > 0.1:
                ec0 = ec + (de - 0.1) / (de - de0) * 100
                return ec0
            de0 = de
        return 250


if __name__ == '__main__':
    import optparse
    parser = optparse.OptionParser(usage='python -m gpaw.atom.optimize '
                                   '[options] element',
                                   description='Optimize dataset')
    parser.add_option('-s', '--summary', action='store_true')
    parser.add_option('-b', '--best', action='store_true')
    parser.add_option('-r', '--run', action='store_true')
    opts, args = parser.parse_args()
    if opts.run:
        symbol = args[0]
        if os.path.isdir(symbol):
            os.chdir(symbol)
            do = DatasetOptimizer(symbol)
        else:
            os.mkdir(symbol)
            os.chdir(symbol)
            do = DatasetOptimizer(symbol)
        do.run()  # *args[1:])
    else:
        if len(args) == 0:
            symbol = os.getcwd().rsplit('/', 1)[1]
            args.append(symbol)
            os.chdir('..')
        for symbol in args:
            os.chdir(symbol)
            do = DatasetOptimizer(symbol)
            if opts.summary:
                do.summary(15)
            elif opts.best:
                do.best1()
            os.chdir('..')
